import json
import os
from unittest.mock import MagicMock, ANY

import pytest

import randovania
from randovania.game_description.resources.pickup_entry import PickupModel, ConditionalResources
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.game import RandovaniaGame
from randovania.games.patchers.randomprime_patcher import RandomprimePatcher, prime1_pickup_details_to_patcher
from randovania.games.prime.patcher_file_lib import pickup_exporter
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.prime1.prime_cosmetic_patches import PrimeCosmeticPatches


@pytest.mark.parametrize("other_player", [False, True])
def test_prime1_pickup_details_to_patcher_shiny_missile(prime1_resource_database, other_player: bool):
    # Setup
    rng = MagicMock()
    rng.randint.return_value = 0
    detail = pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(15),
        scan_text="Your Missile Expansion. Provides 5 Missiles",
        hud_text=["Missile Expansion acquired!"],
        conditional_resources=[ConditionalResources(
            None, None, (
                (prime1_resource_database.get_item_by_name("Missile"), 6),
            ),
        )],
        conversion=[],
        model=PickupModel(RandovaniaGame.METROID_PRIME, "Missile"),
        other_player=other_player,
    )
    if other_player:
        shiny_stuff = {
            'model': 'Missile',
            'scanText': 'Your Missile Expansion. Provides 5 Missiles',
            'hudmemoText': 'Missile Expansion acquired!',
        }
    else:
        shiny_stuff = {
            'model': 'Shiny Missile',
            'scanText': 'Your Shiny Missile Expansion. Provides 5 Missiles',
            'hudmemoText': 'Shiny Missile Expansion acquired!',
        }

    # Run
    result = prime1_pickup_details_to_patcher(detail, False, rng)

    # Assert
    assert result == {
        'type': 'Missile',
        'currIncrease': 6, 'maxIncrease': 6,
        'respawn': False,
        **shiny_stuff,
    }


def test_create_patch_data(test_files_dir):
    # Setup
    file = test_files_dir.joinpath("log_files", "prime1_and_2_multi.rdvgame")
    description = LayoutDescription.from_file(file)
    patcher = RandomprimePatcher()
    players_config = PlayersConfiguration(0, {0: "Prime", 1: "Echoes"})
    cosmetic_patches = PrimeCosmeticPatches()

    # Run
    data = patcher.create_patch_data(description, players_config, cosmetic_patches)

    # Assert
    phendrana = {
        'transports': {
            'Phendrana Drifts North\x00(Phendrana Shorelines)': 'Tallon Overworld East\x00(Frigate Crash Site)',
            'Phendrana Drifts South\x00(Quarantine Cave)': 'Magmoor Caverns West\x00(Monitor Station)'},
        'rooms': {
            'Phendrana Shorelines': {
                'pickups': [
                    {
                        'type': 'Missile',
                        'model': 'Missile',
                        'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                        'hudmemoText': 'Missile Expansion acquired!',
                        'currIncrease': 5, 'maxIncrease': 5,
                        'respawn': False},
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Nothing',
                        'scanText': "Echoes's Light Ammo Expansion. Provides 20 Light Ammo and 1 Item Percentage",
                        'hudmemoText': 'Sent Light Ammo Expansion to Echoes!',
                        'currIncrease': 37, 'maxIncrease': 37,
                        'respawn': False}]},
            'Chozo Ice Temple': {
                'pickups': [
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Energy Tank',
                        'scanText': "Echoes's Energy Tank",
                        'hudmemoText': 'Sent Energy Tank to Echoes!',
                        'currIncrease': 38, 'maxIncrease': 38,
                        'respawn': False}]},
            'Ice Ruins West': {
                'pickups': [
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Missile',
                        'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                        'hudmemoText': 'Sent Missile Expansion to Echoes!',
                        'currIncrease': 39, 'maxIncrease': 39,
                        'respawn': False}]},
            'Ice Ruins East': {
                'pickups': [
                    {
                        'type': 'Missile',
                        'model': 'Missile',
                        'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                        'hudmemoText': 'Missile Expansion acquired!',
                        'currIncrease': 5, 'maxIncrease': 5,
                        'respawn': False},
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Missile',
                        'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                        'hudmemoText': 'Sent Missile Expansion to Echoes!',
                        'currIncrease': 41, 'maxIncrease': 41,
                        'respawn': False}]},
            'Chapel of the Elders': {
                'pickups': [
                    {
                        'type': 'Missile',
                        'model': 'Missile',
                        'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                        'hudmemoText': 'Missile Expansion acquired!',
                        'currIncrease': 5, 'maxIncrease': 5,
                        'respawn': False}]},
            'Ruined Courtyard': {
                'pickups': [
                    {
                        'type': 'Missile',
                        'model': 'Missile',
                        'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                        'hudmemoText': 'Missile Expansion acquired!',
                        'currIncrease': 5, 'maxIncrease': 5,
                        'respawn': False}]},
            'Phendrana Canyon': {
                'pickups': [
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Energy Tank',
                        'scanText': "Echoes's Energy Tank",
                        'hudmemoText': 'Sent Energy Tank to Echoes!',
                        'currIncrease': 44, 'maxIncrease': 44,
                        'respawn': False}]},
            'Quarantine Cave': {
                'pickups': [
                    {
                        'type': 'Energy Tank',
                        'model': 'Energy Tank',
                        'scanText': 'Your Energy Tank',
                        'hudmemoText': 'Energy Tank acquired!',
                        'currIncrease': 1, 'maxIncrease': 1,
                        'respawn': False}]},
            'Research Lab Hydra': {
                'pickups': [
                    {
                        'type': 'Missile',
                        'model': 'Missile',
                        'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                        'hudmemoText': 'Missile Expansion acquired!',
                        'currIncrease': 5, 'maxIncrease': 5,
                        'respawn': False}]},
            'Quarantine Monitor': {
                'pickups': [
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Missile',
                        'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                        'hudmemoText': 'Sent Missile Expansion to Echoes!',
                        'currIncrease': 47, 'maxIncrease': 47,
                        'respawn': False}]},
            'Observatory': {
                'pickups': [
                    {
                        'type': 'Missile',
                        'model': 'Missile',
                        'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                        'hudmemoText': 'Missile Expansion acquired!',
                        'currIncrease': 5, 'maxIncrease': 5,
                        'respawn': False}]},
            'Transport Access': {
                'pickups': [
                    {
                        'type': 'Missile',
                        'model': 'Missile',
                        'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                        'hudmemoText': 'Missile Expansion acquired!',
                        'currIncrease': 5, 'maxIncrease': 5,
                        'respawn': False}]},
            'Control Tower': {
                'pickups': [
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Missile',
                        'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                        'hudmemoText': 'Sent Missile Expansion to Echoes!',
                        'currIncrease': 50, 'maxIncrease': 50,
                        'respawn': False}]},
            'Research Core': {
                'pickups': [
                    {
                        'type': 'Energy Tank',
                        'model': 'Energy Tank',
                        'scanText': 'Your Energy Tank',
                        'hudmemoText': 'Energy Tank acquired!',
                        'currIncrease': 1, 'maxIncrease': 1,
                        'respawn': False}]},
            'Frost Cave': {
                'pickups': [
                    {
                        'type': 'Energy Tank',
                        'model': 'Energy Tank',
                        'scanText': 'Your Energy Tank',
                        'hudmemoText': 'Energy Tank acquired!',
                        'currIncrease': 1, 'maxIncrease': 1,
                        'respawn': False}]},
            'Research Lab Aether': {
                'pickups': [
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Missile',
                        'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                        'hudmemoText': 'Sent Missile Expansion to Echoes!',
                        'currIncrease': 53, 'maxIncrease': 53,
                        'modalHudmemo': True,
                        'respawn': False},
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Power Bomb Expansion',
                        'scanText': "Echoes's Power Bomb Expansion. Provides 1 Power Bomb and 1 Item Percentage",
                        'hudmemoText': 'Sent Power Bomb Expansion to Echoes!',
                        'currIncrease': 54, 'maxIncrease': 54, 'modalHudmemo': True,
                        'respawn': False}]},
            'Gravity Chamber': {
                'pickups': [
                    {
                        'type': 'Missile',
                        'model': 'Missile',
                        'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                        'hudmemoText': 'Missile Expansion acquired!',
                        'currIncrease': 5, 'maxIncrease': 5,
                        'respawn': False},
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Missile',
                        'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                        'hudmemoText': 'Sent Missile Expansion to Echoes!',
                        'currIncrease': 56, 'maxIncrease': 56,
                        'respawn': False}]},
            'Storage Cave': {
                'pickups': [
                    {
                        'type': 'Missile',
                        'model': 'Missile',
                        'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                        'hudmemoText': 'Missile Expansion acquired!',
                        'currIncrease': 5, 'maxIncrease': 5,
                        'respawn': False}]},
            'Security Cave': {
                'pickups': [
                    {
                        'type': 'Unknown Item 1',
                        'model': 'Missile',
                        'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                        'hudmemoText': 'Sent Missile Expansion to Echoes!',
                        'currIncrease': 58, 'maxIncrease': 58,
                        'respawn': False}]}}}
    magmoor = {
        'transports': {
            'Magmoor Caverns North\x00(Lava Lake)': 'Tallon Overworld West\x00(Root Cave)',
            'Magmoor Caverns West\x00(Monitor Station)': 'Phendrana Drifts South\x00(Quarantine Cave)',
            'Magmoor Caverns East\x00(Twin Fires)': 'Phazon Mines East\x00(Main Quarry)',
            'Magmoor Caverns South\x00(Magmoor Workstation, Debris)': 'Chozo Ruins South\x00(Reflecting Pool, Far End)',
            'Magmoor Caverns South\x00(Magmoor Workstation, Save Station)': 'Phazon Mines West\x00(Phazon Processing Center)'},
        'rooms': {
            'Lava Lake': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Morph Ball Bomb', 'scanText': "Echoes's Morph Ball Bomb",
                     'hudmemoText': 'Sent Morph Ball Bomb to Echoes!', 'currIncrease': 91, 'maxIncrease': 91,
                     'respawn': False}]},
            'Triclops Pit': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Power Bomb Expansion',
                     'scanText': "Echoes's Power Bomb Expansion. Provides 1 Power Bomb and 1 Item Percentage",
                     'hudmemoText': 'Sent Power Bomb Expansion to Echoes!',
                     'currIncrease': 92, 'maxIncrease': 92, 'respawn': False}]},
            'Storage Cavern': {
                'pickups': [
                    {'type': 'Energy Tank', 'model': 'Energy Tank',
                     'scanText': 'Your Energy Tank',
                     'hudmemoText': 'Energy Tank acquired!',
                     'currIncrease': 1, 'maxIncrease': 1, 'respawn': False}]},
            'Transport Tunnel A': {
                'pickups': [{'type': 'Missile', 'model': 'Missile',
                             'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                             'hudmemoText': 'Missile Expansion acquired!',
                             'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}]},
            'Warrior Shrine': {
                'pickups': [{'type': 'Unknown Item 1', 'model': 'Nothing',
                             'scanText': "Echoes's Light Ammo Expansion. Provides 20 Light Ammo and 1 Item Percentage",
                             'hudmemoText': 'Sent Light Ammo Expansion to Echoes!',
                             'currIncrease': 95, 'maxIncrease': 95, 'respawn': False}]},
            'Shore Tunnel': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Dark Ammo Expansion. Provides 20 Dark Ammo and 1 Item Percentage",
                     'hudmemoText': 'Sent Dark Ammo Expansion to Echoes!',
                     'currIncrease': 96, 'maxIncrease': 96, 'respawn': False}]},
            'Fiery Shores': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!', 'currIncrease': 97, 'maxIncrease': 97,
                     'respawn': False},
                    {'type': 'Energy Tank', 'model': 'Energy Tank',
                     'scanText': 'Your Energy Tank',
                     'hudmemoText': 'Energy Tank acquired!', 'currIncrease': 1, 'maxIncrease': 1,
                     'respawn': False}
                ]},
            'Plasma Processing': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!', 'currIncrease': 99, 'maxIncrease': 99,
                     'respawn': False}
                ]},
            'Magmoor Workstation': {
                'pickups': [
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!', 'currIncrease': 5, 'maxIncrease': 5,
                     'respawn': False}
                ]
            }
        }}
    mines = {
        'transports': {
            'Phazon Mines East\x00(Main Quarry)': 'Magmoor Caverns East\x00(Twin Fires)',
            'Phazon Mines West\x00(Phazon Processing Center)': 'Magmoor Caverns South\x00(Magmoor Workstation, Save Station)'},
        'rooms': {
            'Main Quarry': {
                'pickups': [
                    {
                        'type': 'Missile', 'model': 'Missile',
                        'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                        'hudmemoText': 'Missile Expansion acquired!', 'currIncrease': 5, 'maxIncrease': 5,
                        'respawn': False}]},
            'Security Access A': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Energy Tank',
                     'scanText': "Echoes's Energy Tank",
                     'hudmemoText': 'Sent Energy Tank to Echoes!', 'currIncrease': 75, 'maxIncrease': 75,
                     'respawn': False}]},
            'Storage Depot B': {
                'pickups': [
                    {'type': 'Energy Tank', 'model': 'Energy Tank',
                     'scanText': 'Your Energy Tank',
                     'hudmemoText': 'Energy Tank acquired!', 'currIncrease': 1, 'maxIncrease': 1,
                     'respawn': False}]},
            'Storage Depot A': {
                'pickups': [
                    {'type': 'Spider Ball', 'model': 'Spider Ball',
                     'scanText': 'Your Spider Ball',
                     'hudmemoText': 'Spider Ball acquired!', 'currIncrease': 1, 'maxIncrease': 1,
                     'respawn': False}]},
            'Elite Research': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Sonic Boom",
                     'hudmemoText': 'Sent Sonic Boom to Echoes!',
                     'currIncrease': 78, 'maxIncrease': 78, 'respawn': False},
                    {'type': 'Nothing', 'model': 'Nothing',
                     'scanText': 'Your Nothing',
                     'hudmemoText': 'Nothing acquired!',
                     'currIncrease': 0, 'maxIncrease': 0, 'respawn': False}
                ]},
            'Elite Control Access': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!',
                     'currIncrease': 80, 'maxIncrease': 80, 'respawn': False}
                ]},
            'Ventilation Shaft': {
                'pickups': [
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}
                ]},
            'Phazon Processing Center': {
                'pickups': [
                    {'type': 'Power Bomb', 'model': 'Power Bomb Expansion',
                     'scanText': 'Your Power Bomb Expansion. Provides 1 Power Bomb',
                     'hudmemoText': 'Power Bomb Expansion acquired!',
                     'currIncrease': 1, 'maxIncrease': 1, 'respawn': False}]},
            'Processing Center Access': {
                'pickups': [
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}]},
            'Elite Quarters': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Light Ammo Expansion. Provides 20 Light Ammo and 1 Item Percentage",
                     'hudmemoText': 'Sent Light Ammo Expansion to Echoes!',
                     'currIncrease': 84, 'maxIncrease': 84, 'respawn': False}]},
            'Central Dynamo': {
                'pickups': [
                    {'type': 'Energy Tank', 'model': 'Energy Tank',
                     'scanText': 'Your Energy Tank',
                     'hudmemoText': 'Energy Tank acquired!', 'currIncrease': 1, 'maxIncrease': 1,
                     'respawn': False}]},
            'Metroid Quarantine B': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Darkburst",
                     'hudmemoText': 'Sent Darkburst to Echoes!',
                     'currIncrease': 86, 'maxIncrease': 86, 'respawn': False}
                ]},
            'Metroid Quarantine A': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!',
                     'currIncrease': 87, 'maxIncrease': 87, 'respawn': False}]},
            'Fungal Hall B': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Light Ammo Expansion. Provides 20 Light Ammo and 1 Item Percentage",
                     'hudmemoText': 'Sent Light Ammo Expansion to Echoes!',
                     'currIncrease': 88, 'maxIncrease': 88, 'respawn': False}]},
            'Phazon Mining Tunnel': {
                'pickups': [
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}
                ]},
            'Fungal Hall Access': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Energy Tank',
                     'scanText': "Echoes's Energy Tank",
                     'hudmemoText': 'Sent Energy Tank to Echoes!',
                     'currIncrease': 90, 'maxIncrease': 90, 'respawn': False}]}}}
    overworld = {
        'transports': {
            'Artifact Temple': 'Crater Entry Point',
            'Tallon Overworld North\x00(Tallon Canyon)': 'Chozo Ruins East\x00(Reflecting Pool, Save Station)',
            'Tallon Overworld East\x00(Frigate Crash Site)': 'Phendrana Drifts North\x00(Phendrana Shorelines)',
            'Tallon Overworld West\x00(Root Cave)': 'Magmoor Caverns North\x00(Lava Lake)',
            'Tallon Overworld South\x00(Great Tree Hall, Upper)': 'Chozo Ruins West\x00(Main Plaza)',
            'Tallon Overworld South\x00(Great Tree Hall, Lower)': 'Chozo Ruins North\x00(Sun Tower)'},
        'rooms': {'Landing Site': {'pickups': [
            {'type': 'Varia Suit', 'model': 'Varia Suit',
             'scanText': 'Your Varia Suit',
             'hudmemoText': 'Varia Suit acquired!', 'currIncrease': 1, 'maxIncrease': 1,
             'respawn': False}]}, 'Alcove': {'pickups': [
            {'type': 'X-Ray Visor', 'model': 'X-Ray Visor',
             'scanText': 'Your X-Ray Visor',
             'hudmemoText': 'X-Ray Visor acquired!', 'currIncrease': 1, 'maxIncrease': 1,
             'respawn': False}]}, 'Frigate Crash Site': {'pickups': [
            {'type': 'Unknown Item 1', 'model': 'Nothing',
             'scanText': "Echoes's Emerald Translator",
             'hudmemoText': 'Sent Emerald Translator to Echoes!',
             'currIncrease': 61, 'maxIncrease': 61, 'respawn': False}]}, 'Overgrown Cavern': {
            'pickups': [{'type': 'Missile', 'model': 'Missile',
                         'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                         'hudmemoText': 'Missile Expansion acquired!',
                         'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}]}, 'Root Cave': {
            'pickups': [{'type': 'Plasma Beam', 'model': 'Plasma Beam',
                         'scanText': 'Your Plasma Beam',
                         'hudmemoText': 'Plasma Beam acquired!',
                         'currIncrease': 1, 'maxIncrease': 1, 'respawn': False}]},
            'Artifact Temple': {'pickups': [
                {'type': 'Unknown Item 1', 'model': 'Nothing',
                 'scanText': "Echoes's Violet Translator",
                 'hudmemoText': 'Sent Violet Translator to Echoes!',
                 'currIncrease': 64, 'maxIncrease': 64, 'respawn': False, 'modalHudmemo': True
                 }]},
            'Transport Tunnel B': {'pickups': [
                {'type': 'Artifact of Wild',
                 'model': 'Artifact of Wild',
                 'scanText': 'Your Artifact of Wild',
                 'hudmemoText': 'Artifact of Wild acquired!',
                 'currIncrease': 1, 'maxIncrease': 1, 'respawn': False}]},
            'Arbor Chamber': {'pickups': [
                {'type': 'Unknown Item 1', 'model': 'Nothing',
                 'scanText': "Echoes's Screw Attack",
                 'hudmemoText': 'Sent Screw Attack to Echoes!',
                 'currIncrease': 66, 'maxIncrease': 66, 'respawn': False}]},
            'Cargo Freight Lift to Deck Gamma': {'pickups': [
                {'type': 'Phazon Suit', 'model': 'Phazon Suit',
                 'scanText': 'Your Phazon Suit',
                 'hudmemoText': 'Phazon Suit acquired!',
                 'currIncrease': 1, 'maxIncrease': 1, 'respawn': False}]},
            'Biohazard Containment': {'pickups': [
                {'type': 'Unknown Item 1', 'model': 'Nothing',
                 'scanText': "Echoes's Dark Ammo Expansion. Provides 20 Dark Ammo and 1 Item Percentage",
                 'hudmemoText': 'Sent Dark Ammo Expansion to Echoes!',
                 'currIncrease': 68, 'maxIncrease': 68, 'respawn': False}]},
            'Hydro Access Tunnel': {'pickups': [
                {'type': 'Missile', 'model': 'Missile',
                 'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                 'hudmemoText': 'Missile Expansion acquired!',
                 'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}]},
            'Great Tree Chamber': {'pickups': [
                {'type': 'Energy Tank', 'model': 'Energy Tank',
                 'scanText': 'Your Energy Tank',
                 'hudmemoText': 'Energy Tank acquired!',
                 'currIncrease': 1, 'maxIncrease': 1, 'respawn': False}]},
            'Life Grove Tunnel': {'pickups': [
                {'type': 'Unknown Item 1', 'model': 'Nothing',
                 'scanText': "Echoes's Light Ammo Expansion. Provides 20 Light Ammo and 1 Item Percentage",
                 'hudmemoText': 'Sent Light Ammo Expansion to Echoes!',
                 'currIncrease': 71, 'maxIncrease': 71, 'respawn': False}]}, 'Life Grove': {
                'pickups': [{'type': 'Boost Ball', 'model': 'Boost Ball',
                             'scanText': 'Your Boost Ball',
                             'hudmemoText': 'Boost Ball acquired!',
                             'currIncrease': 1, 'maxIncrease': 1, 'respawn': False},
                            {'type': 'Missile', 'model': 'Missile',
                             'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                             'hudmemoText': 'Missile Expansion acquired!',
                             'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}]}}}
    chozo = {
        'transports': {
            'Chozo Ruins West\x00(Main Plaza)': 'Tallon Overworld South\x00(Great Tree Hall, Upper)',
            'Chozo Ruins North\x00(Sun Tower)': 'Tallon Overworld South\x00(Great Tree Hall, Lower)',
            'Chozo Ruins East\x00(Reflecting Pool, Save Station)': 'Tallon Overworld North\x00(Tallon Canyon)',
            'Chozo Ruins South\x00(Reflecting Pool, Far End)': 'Magmoor Caverns South\x00(Magmoor Workstation, Debris)'
        },
        'rooms': {
            'Main Plaza': {
                'pickups': [
                    {'type': 'Power Bomb', 'model': 'Power Bomb Expansion',
                     'scanText': 'Your Power Bomb Expansion. Provides 1 Power Bomb',
                     'hudmemoText': 'Power Bomb Expansion acquired!', 'currIncrease': 1, 'maxIncrease': 1,
                     'respawn': False},
                    {'type': 'Power Bomb', 'model': 'Power Bomb',
                     'scanText': 'Your Power Bomb',
                     'hudmemoText': 'Power Bomb acquired!', 'currIncrease': 4, 'maxIncrease': 4,
                     'respawn': False},
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False},
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}
                ]},
            'Ruined Fountain': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}]
            },
            'Ruined Shrine': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Dark Ammo Expansion. Provides 20 Dark Ammo and 1 Item Percentage",
                     'hudmemoText': 'Sent Dark Ammo Expansion to Echoes!',
                     'currIncrease': 6, 'maxIncrease': 6, 'respawn': False, 'modalHudmemo': True},
                    {'type': 'Unknown Item 1', 'model': 'Power Bomb Expansion',
                     'scanText': "Echoes's Power Bomb Expansion. Provides 1 Power Bomb and 1 Item Percentage",
                     'hudmemoText': 'Sent Power Bomb Expansion to Echoes!',
                     'currIncrease': 7, 'maxIncrease': 7, 'respawn': False, 'modalHudmemo': True},
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!',
                     'currIncrease': 8, 'maxIncrease': 8, 'respawn': False, 'modalHudmemo': True}]
            },
            'Vault': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!',
                     'currIncrease': 9, 'maxIncrease': 9, 'respawn': False}
                ]},
            'Training Chamber': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Energy Tank',
                     'scanText': "Echoes's Energy Tank",
                     'hudmemoText': 'Sent Energy Tank to Echoes!',
                     'currIncrease': 10, 'maxIncrease': 10, 'respawn': False}]},
            'Ruined Nursery': {
                'pickups': [
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}
                ]},
            'Training Chamber Access': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!',
                     'currIncrease': 12, 'maxIncrease': 12, 'respawn': False}]},
            'Magma Pool': {
                'pickups': [
                    {'type': 'Energy Tank', 'model': 'Energy Tank',
                     'scanText': 'Your Energy Tank',
                     'hudmemoText': 'Energy Tank acquired!',
                     'currIncrease': 1, 'maxIncrease': 1, 'respawn': False}]},
            'Tower of Light': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!',
                     'currIncrease': 14, 'maxIncrease': 14, 'respawn': False
                     }]},
            'Tower Chamber': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Power Bomb Expansion',
                     'scanText': "Echoes's Power Bomb Expansion. Provides 1 Power Bomb and 1 Item Percentage",
                     'hudmemoText': 'Sent Power Bomb Expansion to Echoes!',
                     'currIncrease': 15, 'maxIncrease': 15, 'respawn': False}]},
            'Ruined Gallery': {
                'pickups': [
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False},
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!',
                     'currIncrease': 17, 'maxIncrease': 17, 'respawn': False}
                ]},
            'Transport Access North': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!',
                     'currIncrease': 18, 'maxIncrease': 18, 'respawn': False, 'modalHudmemo': True
                     }]},
            'Gathering Hall': {
                'pickups': [
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}]},
            'Hive Totem': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Dark Ammo Expansion. Provides 20 Dark Ammo and 1 Item Percentage",
                     'hudmemoText': 'Sent Dark Ammo Expansion to Echoes!',
                     'currIncrease': 20, 'maxIncrease': 20, 'respawn': False, 'modalHudmemo': True,
                     }]},
            'Sunchamber': {
                'pickups': [
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False},
                    {'type': 'Wavebuster', 'model': 'Wavebuster',
                     'scanText': 'Your Wavebuster',
                     'hudmemoText': 'Wavebuster acquired!',
                     'currIncrease': 1, 'maxIncrease': 1, 'respawn': False}]},
            'Watery Hall Access': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Dark Ammo Expansion. Provides 20 Dark Ammo and 1 Item Percentage",
                     'hudmemoText': 'Sent Dark Ammo Expansion to Echoes!',
                     'currIncrease': 23, 'maxIncrease': 23, 'respawn': False}]},
            'Watery Hall': {
                'pickups': [
                    {'type': 'Energy Tank', 'model': 'Energy Tank',
                     'scanText': 'Your Energy Tank',
                     'hudmemoText': 'Energy Tank acquired!',
                     'currIncrease': 1, 'maxIncrease': 1, 'respawn': False},
                    {'type': 'Energy Tank', 'model': 'Energy Tank',
                     'scanText': 'Your Energy Tank',
                     'hudmemoText': 'Energy Tank acquired!',
                     'currIncrease': 1, 'maxIncrease': 1, 'respawn': False}]},
            'Dynamo': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Dark Ammo Expansion. Provides 20 Dark Ammo and 1 Item Percentage",
                     'hudmemoText': 'Sent Dark Ammo Expansion to Echoes!',
                     'currIncrease': 26, 'maxIncrease': 26, 'respawn': False},
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False}]
            },
            'Burn Dome': {
                'pickups': [
                    {'type': 'Missile', 'model': 'Missile',
                     'scanText': 'Your Missile Expansion. Provides 5 Missiles',
                     'hudmemoText': 'Missile Expansion acquired!',
                     'currIncrease': 5, 'maxIncrease': 5, 'respawn': False},
                    {'type': 'Unknown Item 1', 'model': 'Energy Tank',
                     'scanText': "Echoes's Energy Tank",
                     'hudmemoText': 'Sent Energy Tank to Echoes!',
                     'currIncrease': 29, 'maxIncrease': 29, 'respawn': False}]},
            'Furnace': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Light Ammo Expansion. Provides 20 Light Ammo and 1 Item Percentage",
                     'hudmemoText': 'Sent Light Ammo Expansion to Echoes!',
                     'currIncrease': 30, 'maxIncrease': 30, 'respawn': False},
                    {'type': 'Unknown Item 1', 'model': 'Energy Tank',
                     'scanText': "Echoes's Energy Tank",
                     'hudmemoText': 'Sent Energy Tank to Echoes!',
                     'currIncrease': 31, 'maxIncrease': 31, 'respawn': False}]},
            'Hall of the Elders': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Missile',
                     'scanText': "Echoes's Missile Expansion. Provides 5 Missiles and 1 Item Percentage",
                     'hudmemoText': 'Sent Missile Expansion to Echoes!',
                     'currIncrease': 32, 'maxIncrease': 32, 'respawn': False}]}, 'Crossway': {
                'pickups': [
                    {'type': 'Unknown Item 1', 'model': 'Nothing',
                     'scanText': "Echoes's Dark Ammo Expansion. Provides 20 Dark Ammo and 1 Item Percentage",
                     'hudmemoText': 'Sent Dark Ammo Expansion to Echoes!',
                     'currIncrease': 33, 'maxIncrease': 33, 'respawn': False}]},
            'Elder Chamber': {'pickups': [
                {'type': 'Unknown Item 1', 'model': 'Energy Tank',
                 'scanText': "Echoes's Energy Tank",
                 'hudmemoText': 'Sent Energy Tank to Echoes!',
                 'currIncrease': 34, 'maxIncrease': 34, 'respawn': False}]},
            'Antechamber': {
                'pickups': [{
                    'type': 'Unknown Item 1',
                    'model': 'Power Bomb Expansion',
                    'scanText': "Echoes's Power Bomb Expansion. Provides 1 Power Bomb and 1 Item Percentage",
                    'hudmemoText': 'Sent Power Bomb Expansion to Echoes!',
                    'currIncrease': 35, 'maxIncrease': 35, 'respawn': False
                }]}}}

    expected_data = {
        'seed': 1499122484,
        'hasSpoiler': True,
        'preferences': {
            'artifactHintBehavior': None,
            'automaticCrashScreen': True,
            'mapDefaultState': 'default',
            'qolCosmetic': True,
            'qolGameBreaking': True,
            'qolCutscenes': 'original',
            'qolPickupScans': False,
            'quickplay': False,
            'quiet': False,
            'trilogyDiscPath': None,
        },
        'gameConfig': {
            'artifactHints': {
                'Artifact of Chozo': '&push;&main-color=#c300ff;Artifact of Chozo&pop; is located in '
                                     "&push;&main-color=#d4cc33;Echoes&pop;'s "
                                     "&push;&main-color=#89a1ff;Dark Agon Wastes - Ing Cache 1&pop;.",
                'Artifact of Elder': '&push;&main-color=#c300ff;Artifact of Elder&pop; is located in '
                                     "&push;&main-color=#d4cc33;Echoes&pop;'s "
                                     '&push;&main-color=#89a1ff;Agon Wastes - Sand Cache&pop;.',
                'Artifact of Lifegiver': '&push;&main-color=#c300ff;Artifact of Lifegiver&pop; is located in '
                                         "&push;&main-color=#d4cc33;Echoes&pop;'s "
                                         '&push;&main-color=#89a1ff;Temple Grounds - Storage Cavern B&pop;.',
                'Artifact of Nature': '&push;&main-color=#c300ff;Artifact of Nature&pop; is located in '
                                      "&push;&main-color=#d4cc33;Echoes&pop;'s "
                                      '&push;&main-color=#89a1ff;Sanctuary Fortress - Hall of Combat Mastery&pop;.',
                'Artifact of Newborn': '&push;&main-color=#c300ff;Artifact of Newborn&pop; is located in '
                                       "&push;&main-color=#d4cc33;Echoes&pop;'s "
                                       '&push;&main-color=#89a1ff;Great Temple - Transport A Access&pop;.',
                'Artifact of Spirit': '&push;&main-color=#c300ff;Artifact of Spirit&pop; is located in '
                                      "&push;&main-color=#d4cc33;Echoes&pop;'s "
                                      '&push;&main-color=#89a1ff;Temple Grounds - Transport to Agon Wastes&pop;.',
                'Artifact of Strength': '&push;&main-color=#c300ff;Artifact of Strength&pop; is located in '
                                        "&push;&main-color=#d4cc33;Echoes&pop;'s "
                                        '&push;&main-color=#89a1ff;Agon Wastes - Storage D&pop;.',
                'Artifact of Sun': '&push;&main-color=#c300ff;Artifact of Sun&pop; has no need to be located.',
                'Artifact of Truth': '&push;&main-color=#c300ff;Artifact of Truth&pop; is located in '
                                     "&push;&main-color=#d4cc33;Echoes&pop;'s "
                                     '&push;&main-color=#89a1ff;Agon Wastes - Mining Station Access&pop;.',
                'Artifact of Warrior': '&push;&main-color=#c300ff;Artifact of Warrior&pop; is located in '
                                       "&push;&main-color=#d4cc33;Echoes&pop;'s "
                                       '&push;&main-color=#89a1ff;Agon Wastes - Portal Access A&pop;.',
                'Artifact of Wild': '&push;&main-color=#c300ff;Artifact of Wild&pop; is located in '
                                    "&push;&main-color=#d4cc33;Prime&pop;'s "
                                    '&push;&main-color=#89a1ff;Tallon Overworld - Transport Tunnel B&pop;.',
                'Artifact of World': '&push;&main-color=#c300ff;Artifact of World&pop; is located in '
                                     "&push;&main-color=#d4cc33;Echoes&pop;'s "
                                     '&push;&main-color=#89a1ff;Ing Hive - Culling Chamber&pop;.'
            },

            'artifactTempleLayerOverrides': {
                'Artifact of Chozo': True,
                'Artifact of Elder': True,
                'Artifact of Lifegiver': True,
                'Artifact of Nature': True,
                'Artifact of Newborn': True,
                'Artifact of Spirit': True,
                'Artifact of Strength': True,
                'Artifact of Sun': False,
                'Artifact of Truth': True,
                'Artifact of Warrior': True,
                'Artifact of Wild': True,
                'Artifact of World': True,
            },
            'autoEnabledElevators': False,
            'multiworldDolPatches': True,
            'creditsString': (
                "&push;&font=C29C51F1;&main-color=#89D6FF;Major Item Locations&pop;"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Charge Beam&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Sanctuary Fortress - Sanctuary Energy Controller"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Wave Beam&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Torvus Bog - Great Bridge"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Ice Beam&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Temple Grounds - Temple Assembly Site"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Plasma Beam&pop;\n"
                "&push;&main-color=#d4cc33;Prime&pop;'s Tallon Overworld - Root Cave"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Grapple Beam&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Dark Agon Wastes - Junction Site"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Thermal Visor&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Torvus Bog - Torvus Energy Controller"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;X-Ray Visor&pop;\n"
                "&push;&main-color=#d4cc33;Prime&pop;'s Tallon Overworld - Alcove"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Boost Ball&pop;\n"
                "&push;&main-color=#d4cc33;Prime&pop;'s Tallon Overworld - Life Grove"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Spider Ball&pop;\n"
                "&push;&main-color=#d4cc33;Prime&pop;'s Phazon Mines - Storage Depot A"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Power Bomb&pop;\n"
                "&push;&main-color=#d4cc33;Prime&pop;'s Chozo Ruins - Main Plaza"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Varia Suit&pop;\n"
                "&push;&main-color=#d4cc33;Prime&pop;'s Tallon Overworld - Landing Site"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Gravity Suit&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Great Temple - Transport B Access"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Phazon Suit&pop;\n"
                "&push;&main-color=#d4cc33;Prime&pop;'s Tallon Overworld - Cargo Freight Lift to Deck Gamma"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Super Missile&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Temple Grounds - Windchamber Gateway"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Wavebuster&pop;\n"
                "&push;&main-color=#d4cc33;Prime&pop;'s Chozo Ruins - Sunchamber"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Ice Spreader&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Agon Wastes - Sandcanyon"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Flamethrower&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Torvus Bog - Training Chamber"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of Chozo&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Dark Agon Wastes - Ing Cache 1"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of Elder&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Agon Wastes - Sand Cache"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of Lifegiver&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Temple Grounds - Storage Cavern B"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of Nature&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Sanctuary Fortress - Hall of Combat Mastery"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of Newborn&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Great Temple - Transport A Access"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of Spirit&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Temple Grounds - Transport to Agon Wastes"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of Strength&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Agon Wastes - Storage D"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of Truth&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Agon Wastes - Mining Station Access"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of Warrior&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Agon Wastes - Portal Access A"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of Wild&pop;\n"
                "&push;&main-color=#d4cc33;Prime&pop;'s Tallon Overworld - Transport Tunnel B"

                "\n\n&push;&font=C29C51F1;&main-color=#33ffd6;Artifact of World&pop;\n"
                "&push;&main-color=#d4cc33;Echoes&pop;'s Ing Hive - Culling Chamber"
            ),
            'etankCapacity': 110,

            "mainPlazaDoor": True,
            "backwardsFrigate": True,
            "backwardsLabs": True,
            "backwardsUpperMines": True,
            "backwardsLowerMines": False,
            "phazonEliteWithoutDynamo": True,

            'gameBanner': {'description': 'Seed Hash: Caretaker Lumigek Skiff',
                           'gameName': 'Metroid Prime: Randomizer',
                           'gameNameFull': 'Metroid Prime: Randomizer - Z2OYS7HX'},
            'heatDamagePerSec': 10.0,
            'itemMaxCapacity': {'Unknown Item 1': 65536},
            'mainMenuMessage': f'Randovania v{randovania.VERSION}\nCaretaker Lumigek Skiff',
            'nonvariaHeatDamage': True,
            'staggeredSuitDamage': True,
            'startingItems': {
                'bombs': False,
                'boostBall': False,
                'combatVisor': True,
                'charge': False,
                'energyTanks': 0,
                'flamethrower': False,
                'grapple': False,
                'gravitySuit': False,
                'ice': False,
                'iceSpreader': False,
                'missiles': 5,
                'morphBall': True,
                'phazonSuit': False,
                'plasma': False,
                'powerBeam': True,
                'powerBombs': 0,
                'scanVisor': True,
                'spaceJump': True,
                'spiderBall': False,
                'superMissile': False,
                'thermalVisor': False,
                'variaSuit': False,
                'wave': False,
                'wavebuster': False,
                'xray': False,
            },
            'startingMemo': 'Artifact of Sun, 5 Missiles',
            'startingRoom': 'Tallon Overworld:Landing Site',
            'warpToStart': False,
        },
        "tweaks": {},
        'levelData': {
            'Impact Crater': {
                'rooms': {},
                'transports': {
                    'Crater Entry Point': 'Artifact Temple',
                },
            },
            'Phendrana Drifts': phendrana,
            'Frigate Orpheon': {'transports': {}, 'rooms': {}},
            'Magmoor Caverns': magmoor,
            'Phazon Mines': mines,
            'Tallon Overworld': overworld,
            'Chozo Ruins': chozo
        },
    }

    assert data == expected_data

def test_patch_game(mocker, tmp_path):
    mock_symbols_for_file: MagicMock = mocker.patch("py_randomprime.symbols_for_file", return_value={
        "UpdateHintState__13CStateManagerFf": 0x80044D38,
    })
    mock_patch_iso_raw: MagicMock = mocker.patch("py_randomprime.patch_iso_raw")
    patch_data = {"patch": "data", 'gameConfig': {}, 'hasSpoiler': True}
    game_files_path = MagicMock()
    progress_update = MagicMock()

    patcher = RandomprimePatcher()

    # Run
    patcher.patch_game(tmp_path.joinpath("input.iso"), tmp_path.joinpath("output.iso"),
                       patch_data, game_files_path, progress_update)

    # Assert
    expected = {
        "patch": "data",
        'gameConfig': {
            "updateHintStateReplacement": [
                148, 33, 255, 204, 124, 8, 2, 166, 144, 1, 0, 56, 191, 193, 0, 44, 124, 127, 27, 120, 136, 159, 0, 2,
                44, 4, 0, 0, 64, 130, 0, 24, 187, 193, 0, 44, 128, 1, 0, 56, 124, 8, 3, 166, 56, 33, 0, 52, 78,
                128, 0, 32, 56, 192, 0, 0, 152, 223, 0, 2, 63, 192, 128, 4, 99, 222, 77, 148, 56, 128, 1, 0, 124, 4,
                247, 172, 44, 4, 0, 0, 56, 132, 255, 224, 64, 130, 255, 244, 124, 0, 4, 172, 76, 0, 1, 44, 187,
                193, 0, 44, 128, 1, 0, 56, 124, 8, 3, 166, 56, 33, 0, 52, 78, 128, 0, 32
            ]
        },
        "inputIso": os.fspath(tmp_path.joinpath("input.iso")),
        "outputIso": os.fspath(tmp_path.joinpath("output.iso")),
    }
    mock_symbols_for_file.assert_called_once_with(tmp_path.joinpath("input.iso"))
    mock_patch_iso_raw.assert_called_once_with(json.dumps(expected, indent=4, separators=(',', ': ')), ANY)



import functools
from typing import Callable, TypeVar

from PySide2 import QtWidgets

T = TypeVar("T")


def on_checked(widget: QtWidgets.QCheckBox, call: Callable[[bool], None]):
    @functools.wraps(call)
    def wrapper(value: int):
        return call(bool(value))

    widget.stateChanged.connect(wrapper)


def combo_set_to_value(combo: QtWidgets.QComboBox, item):
    combo.setCurrentIndex(combo.findData(item))


def on_combo(widget: QtWidgets.QCheckBox, call: Callable[[T], None]):
    @functools.wraps(call)
    def wrapper():
        return call(widget.currentData())

    widget.currentIndexChanged.connect(wrapper)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_logic_damage.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetLogicDamage(object):
    def setupUi(self, PresetLogicDamage):
        if not PresetLogicDamage.objectName():
            PresetLogicDamage.setObjectName(u"PresetLogicDamage")
        PresetLogicDamage.resize(505, 463)
        self.centralWidget = QWidget(PresetLogicDamage)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(2, 2, 2, 2)
        self.damage_strictness_group = QGroupBox(self.centralWidget)
        self.damage_strictness_group.setObjectName(u"damage_strictness_group")
        self.damage_strictness_layout = QVBoxLayout(self.damage_strictness_group)
        self.damage_strictness_layout.setSpacing(6)
        self.damage_strictness_layout.setContentsMargins(11, 11, 11, 11)
        self.damage_strictness_layout.setObjectName(u"damage_strictness_layout")
        self.damage_strictness_label = QLabel(self.damage_strictness_group)
        self.damage_strictness_label.setObjectName(u"damage_strictness_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.damage_strictness_label.sizePolicy().hasHeightForWidth())
        self.damage_strictness_label.setSizePolicy(sizePolicy)
        self.damage_strictness_label.setWordWrap(True)

        self.damage_strictness_layout.addWidget(self.damage_strictness_label)

        self.damage_strictness_spacer = QSpacerItem(20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.damage_strictness_layout.addItem(self.damage_strictness_spacer)

        self.damage_strictness_combo = QComboBox(self.damage_strictness_group)
        self.damage_strictness_combo.addItem("")
        self.damage_strictness_combo.addItem("")
        self.damage_strictness_combo.addItem("")
        self.damage_strictness_combo.setObjectName(u"damage_strictness_combo")

        self.damage_strictness_layout.addWidget(self.damage_strictness_combo)


        self.verticalLayout.addWidget(self.damage_strictness_group)

        self.damage_strictness_spacer_2 = QSpacerItem(20, 302, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.damage_strictness_spacer_2)

        PresetLogicDamage.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetLogicDamage)

        QMetaObject.connectSlotsByName(PresetLogicDamage)
    # setupUi

    def retranslateUi(self, PresetLogicDamage):
        PresetLogicDamage.setWindowTitle(QCoreApplication.translate("PresetLogicDamage", u"Damage", None))
        self.damage_strictness_group.setTitle(QCoreApplication.translate("PresetLogicDamage", u"Logic damage strictness", None))
        self.damage_strictness_label.setText(QCoreApplication.translate("PresetLogicDamage", u"<html><head/><body><p>Certain locations, such as rooms without safe zones in Dark Aether or bosses, requires a certain number of energy tanks (or suits).</p><p>This setting controls how much energy the logic will expect you to have to reach these locations.</p></body></html>", None))
        self.damage_strictness_combo.setItemText(0, QCoreApplication.translate("PresetLogicDamage", u"Strict (1\u00d7)", None))
        self.damage_strictness_combo.setItemText(1, QCoreApplication.translate("PresetLogicDamage", u"Medium (1.5\u00d7)", None))
        self.damage_strictness_combo.setItemText(2, QCoreApplication.translate("PresetLogicDamage", u"Lenient (2\u00d7)", None))

        self.damage_strictness_combo.setCurrentText(QCoreApplication.translate("PresetLogicDamage", u"Strict (1\u00d7)", None))
    # retranslateUi




import contextlib
from dataclasses import dataclass
from typing import TypeVar, Type
from unittest.mock import patch, MagicMock

import pytest

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder, BitPackValue
from randovania.games.game import RandovaniaGame
from randovania.layout.base.ammo_configuration import AmmoConfiguration
from randovania.layout.base.available_locations import AvailableLocationsConfiguration
from randovania.layout.base.base_configuration import StartingLocationList
from randovania.layout.prime2.beam_configuration import BeamConfiguration
from randovania.layout.base.damage_strictness import LayoutDamageStrictness
from randovania.layout.prime2.echoes_configuration import EchoesConfiguration, LayoutSkyTempleKeyMode
from randovania.layout.prime2.hint_configuration import HintConfiguration
from randovania.layout.base.major_items_configuration import MajorItemsConfiguration
from randovania.layout.base.pickup_model import PickupModelStyle
from randovania.layout.lib.teleporters import TeleporterConfiguration
from randovania.layout.prime2.translator_configuration import TranslatorConfiguration
from randovania.layout.base.trick_level_configuration import TrickLevelConfiguration

T = TypeVar("T")


@dataclass(frozen=True)
class DummyValue(BitPackValue):
    def bit_pack_encode(self, metadata):
        yield from []

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata):
        raise cls()


def empty_bit_pack_encode(*args):
    yield from []


@contextlib.contextmanager
def make_dummy(cls: Type[T]) -> T:
    m = MagicMock(spec=cls)
    m.bit_pack_encode = empty_bit_pack_encode
    m.as_json = m

    ret = MagicMock(return_value=m)
    with patch.multiple(cls, from_json=ret, bit_pack_unpack=ret, as_json=ret):
        yield m


@pytest.fixture(
    params=[
        {"encoded": b'@X\xc7\xf1p\xc0K\x03\xfe',
         "sky_temple_keys": LayoutSkyTempleKeyMode.NINE.value,
         },
        {"encoded": b'@\x00\xc7\xf1p\xc0K\x03\xfe',
         "sky_temple_keys": LayoutSkyTempleKeyMode.ALL_BOSSES.value,
         },
        {"encoded": b'@"/\xf1p\xc0K\x03\xfe',
         "sky_temple_keys": LayoutSkyTempleKeyMode.TWO.value,
         "energy_per_tank": 280,
         },
        {"encoded": b'@\x08\xc7\xf4^\xc0K\x03\xfe',
         "sky_temple_keys": LayoutSkyTempleKeyMode.ALL_GUARDIANS.value,
         "varia_suit_damage": 18.0,
         },
        {"encoded": b'\x10\x08\xc7\xf1p\xc0K\x03\xfe',
         "pickup_model_style": PickupModelStyle.HIDE_MODEL.value,
         "sky_temple_keys": LayoutSkyTempleKeyMode.ALL_GUARDIANS.value,
         "damage_strictness": LayoutDamageStrictness.STRICT.value,
         },
    ],
    name="layout_config_with_data")
def _layout_config_with_data(request, default_layout_configuration):
    data = default_layout_configuration.as_json
    for key, value in request.param.items():
        if key != "encoded":
            assert key in data
            data[key] = value

    types_to_mock = {
        "trick_level": TrickLevelConfiguration,
        "starting_location": StartingLocationList,
        "available_locations": AvailableLocationsConfiguration,
        "major_items_configuration": MajorItemsConfiguration,
        "ammo_configuration": AmmoConfiguration,

        "elevators": TeleporterConfiguration,
        "translator_configuration": TranslatorConfiguration,
        "hints": HintConfiguration,
        "beam_configuration": BeamConfiguration,
    }

    with contextlib.ExitStack() as stack:
        for key, cls in types_to_mock.items():
            assert key in data
            data[key] = stack.enter_context(make_dummy(cls))
        yield request.param["encoded"], EchoesConfiguration.from_json(data, game=RandovaniaGame.METROID_PRIME_ECHOES), data


def test_decode(layout_config_with_data):
    # Setup
    data, expected, _ = layout_config_with_data

    # Run
    decoder = BitPackDecoder(data)
    result = EchoesConfiguration.bit_pack_unpack(decoder, {})

    # Assert
    assert result == expected


def test_encode(layout_config_with_data):
    # Setup
    expected, value, expected_json = layout_config_with_data

    # Run
    result = bitpacking.pack_value(value)
    as_json = value.as_json

    # Assert
    assert result == expected
    assert as_json == expected_json



import aiohttp
import pytest
from unittest.mock import AsyncMock, MagicMock
from randovania.interface_common import github_releases_data

@pytest.mark.asyncio
async def test_download_from_github_success(mocker):
    # Setup
    mock_get_response: MagicMock = mocker.patch.object(aiohttp.ClientSession, 'get')
    mock_get_response.return_value.__aenter__.return_value.json = AsyncMock()
 
    # Run
    returned_value = await github_releases_data._download_from_github()
    
    # Assert
    mock_get_response.assert_called()
    mock_get_response.return_value.__aenter__.return_value.json.assert_awaited()
    assert returned_value == mock_get_response.return_value.__aenter__.return_value.json.return_value


@pytest.mark.asyncio
async def test_download_from_github_bad_response_is_caught(mocker):
    # Setup
    mock_get_response: MagicMock = mocker.patch.object(aiohttp.ClientSession, 'get')
    mock_get_response.return_value.__aenter__.return_value.raise_for_status = MagicMock(side_effect = aiohttp.ClientResponseError(None, None))
 
    # Run
    returned_value = await github_releases_data._download_from_github()
    
    # Assert
    mock_get_response.assert_called()
    mock_get_response.return_value.__aenter__.return_value.raise_for_status.assert_called()
    assert returned_value == None


@pytest.mark.asyncio
async def test_download_from_github_connection_failure_is_caught(mocker):
    # Setup
    mock_get_response: MagicMock = mocker.patch.object(aiohttp.ClientSession, 'get', side_effect=aiohttp.ClientConnectionError())

    # Run
    returned_value = await github_releases_data._download_from_github()
    
    # Assert
    mock_get_response.assert_called()
    assert returned_value == None



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_patcher_energy.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetPatcherEnergy(object):
    def setupUi(self, PresetPatcherEnergy):
        if not PresetPatcherEnergy.objectName():
            PresetPatcherEnergy.setObjectName(u"PresetPatcherEnergy")
        PresetPatcherEnergy.resize(476, 628)
        self.centralWidget = QWidget(PresetPatcherEnergy)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.main_layout = QVBoxLayout(self.centralWidget)
        self.main_layout.setSpacing(6)
        self.main_layout.setContentsMargins(11, 11, 11, 11)
        self.main_layout.setObjectName(u"main_layout")
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.scroll_area = QScrollArea(self.centralWidget)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, -5, 460, 648))
        sizePolicy = QSizePolicy(QSizePolicy.Ignored, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scroll_area_contents.sizePolicy().hasHeightForWidth())
        self.scroll_area_contents.setSizePolicy(sizePolicy)
        self.scroll_area_layout = QVBoxLayout(self.scroll_area_contents)
        self.scroll_area_layout.setSpacing(6)
        self.scroll_area_layout.setContentsMargins(11, 11, 11, 11)
        self.scroll_area_layout.setObjectName(u"scroll_area_layout")
        self.scroll_area_layout.setContentsMargins(1, 1, 1, 0)
        self.energy_tank_box = QGroupBox(self.scroll_area_contents)
        self.energy_tank_box.setObjectName(u"energy_tank_box")
        self.gridLayout_2 = QGridLayout(self.energy_tank_box)
        self.gridLayout_2.setSpacing(6)
        self.gridLayout_2.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.dangerous_tank_check = QCheckBox(self.energy_tank_box)
        self.dangerous_tank_check.setObjectName(u"dangerous_tank_check")

        self.gridLayout_2.addWidget(self.dangerous_tank_check, 4, 0, 1, 2)

        self.energy_tank_capacity_label = QLabel(self.energy_tank_box)
        self.energy_tank_capacity_label.setObjectName(u"energy_tank_capacity_label")

        self.gridLayout_2.addWidget(self.energy_tank_capacity_label, 2, 0, 1, 1)

        self.progressive_damage_reduction_check = QCheckBox(self.energy_tank_box)
        self.progressive_damage_reduction_check.setObjectName(u"progressive_damage_reduction_check")

        self.gridLayout_2.addWidget(self.progressive_damage_reduction_check, 5, 0, 1, 2)

        self.energy_tank_capacity_spin_box = QSpinBox(self.energy_tank_box)
        self.energy_tank_capacity_spin_box.setObjectName(u"energy_tank_capacity_spin_box")
        self.energy_tank_capacity_spin_box.setMinimum(2)
        self.energy_tank_capacity_spin_box.setMaximum(1000)

        self.gridLayout_2.addWidget(self.energy_tank_capacity_spin_box, 2, 1, 1, 1)

        self.energy_tank_capacity_description = QLabel(self.energy_tank_box)
        self.energy_tank_capacity_description.setObjectName(u"energy_tank_capacity_description")
        self.energy_tank_capacity_description.setWordWrap(True)

        self.gridLayout_2.addWidget(self.energy_tank_capacity_description, 0, 0, 1, 2)

        self.line = QFrame(self.energy_tank_box)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.gridLayout_2.addWidget(self.line, 3, 0, 1, 2)

        self.energy_tank_capacity_spacer = QSpacerItem(20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.gridLayout_2.addItem(self.energy_tank_capacity_spacer, 1, 0, 1, 2)


        self.scroll_area_layout.addWidget(self.energy_tank_box)

        self.safe_zone_box = QGroupBox(self.scroll_area_contents)
        self.safe_zone_box.setObjectName(u"safe_zone_box")
        self.gridLayout = QGridLayout(self.safe_zone_box)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.safe_zone_regen_spin = QDoubleSpinBox(self.safe_zone_box)
        self.safe_zone_regen_spin.setObjectName(u"safe_zone_regen_spin")
        self.safe_zone_regen_spin.setMaximum(100.000000000000000)
        self.safe_zone_regen_spin.setSingleStep(0.100000000000000)
        self.safe_zone_regen_spin.setValue(1.000000000000000)

        self.gridLayout.addWidget(self.safe_zone_regen_spin, 2, 1, 1, 1)

        self.safe_zone_regen_label = QLabel(self.safe_zone_box)
        self.safe_zone_regen_label.setObjectName(u"safe_zone_regen_label")

        self.gridLayout.addWidget(self.safe_zone_regen_label, 2, 0, 1, 1)

        self.safe_zone_logic_heal_check = QCheckBox(self.safe_zone_box)
        self.safe_zone_logic_heal_check.setObjectName(u"safe_zone_logic_heal_check")
        self.safe_zone_logic_heal_check.setEnabled(False)
        self.safe_zone_logic_heal_check.setChecked(True)

        self.gridLayout.addWidget(self.safe_zone_logic_heal_check, 1, 0, 1, 2)

        self.safe_zone_description = QLabel(self.safe_zone_box)
        self.safe_zone_description.setObjectName(u"safe_zone_description")

        self.gridLayout.addWidget(self.safe_zone_description, 0, 0, 1, 2)


        self.scroll_area_layout.addWidget(self.safe_zone_box)

        self.dark_aether_box = QGroupBox(self.scroll_area_contents)
        self.dark_aether_box.setObjectName(u"dark_aether_box")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.dark_aether_box.sizePolicy().hasHeightForWidth())
        self.dark_aether_box.setSizePolicy(sizePolicy1)
        self.dark_aether_layout_2 = QGridLayout(self.dark_aether_box)
        self.dark_aether_layout_2.setSpacing(6)
        self.dark_aether_layout_2.setContentsMargins(11, 11, 11, 11)
        self.dark_aether_layout_2.setObjectName(u"dark_aether_layout_2")
        self.varia_suit_spin_box = QDoubleSpinBox(self.dark_aether_box)
        self.varia_suit_spin_box.setObjectName(u"varia_suit_spin_box")
        self.varia_suit_spin_box.setMaximum(60.000000000000000)
        self.varia_suit_spin_box.setStepType(QAbstractSpinBox.AdaptiveDecimalStepType)
        self.varia_suit_spin_box.setValue(6.000000000000000)

        self.dark_aether_layout_2.addWidget(self.varia_suit_spin_box, 2, 1, 1, 1)

        self.varia_suit_label = QLabel(self.dark_aether_box)
        self.varia_suit_label.setObjectName(u"varia_suit_label")

        self.dark_aether_layout_2.addWidget(self.varia_suit_label, 2, 0, 1, 1)

        self.dark_suit_label = QLabel(self.dark_aether_box)
        self.dark_suit_label.setObjectName(u"dark_suit_label")

        self.dark_aether_layout_2.addWidget(self.dark_suit_label, 3, 0, 1, 1)

        self.dark_suit_spin_box = QDoubleSpinBox(self.dark_aether_box)
        self.dark_suit_spin_box.setObjectName(u"dark_suit_spin_box")
        self.dark_suit_spin_box.setMaximum(60.000000000000000)
        self.dark_suit_spin_box.setStepType(QAbstractSpinBox.AdaptiveDecimalStepType)
        self.dark_suit_spin_box.setValue(1.200000000000000)

        self.dark_aether_layout_2.addWidget(self.dark_suit_spin_box, 3, 1, 1, 1)

        self.dark_aether_label = QLabel(self.dark_aether_box)
        self.dark_aether_label.setObjectName(u"dark_aether_label")
        sizePolicy2 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.dark_aether_label.sizePolicy().hasHeightForWidth())
        self.dark_aether_label.setSizePolicy(sizePolicy2)
        self.dark_aether_label.setWordWrap(False)

        self.dark_aether_layout_2.addWidget(self.dark_aether_label, 0, 0, 1, 2)

        self.dark_aether_box_spacer = QSpacerItem(20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.dark_aether_layout_2.addItem(self.dark_aether_box_spacer, 1, 0, 1, 2)


        self.scroll_area_layout.addWidget(self.dark_aether_box)

        self.heated_damage_box = QGroupBox(self.scroll_area_contents)
        self.heated_damage_box.setObjectName(u"heated_damage_box")
        self.gridLayout_3 = QGridLayout(self.heated_damage_box)
        self.gridLayout_3.setSpacing(6)
        self.gridLayout_3.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.heated_damage_description = QLabel(self.heated_damage_box)
        self.heated_damage_description.setObjectName(u"heated_damage_description")

        self.gridLayout_3.addWidget(self.heated_damage_description, 0, 0, 1, 2)

        self.heated_damage_varia_check = QCheckBox(self.heated_damage_box)
        self.heated_damage_varia_check.setObjectName(u"heated_damage_varia_check")
        self.heated_damage_varia_check.setChecked(True)

        self.gridLayout_3.addWidget(self.heated_damage_varia_check, 1, 0, 1, 2)

        self.heated_damage_label = QLabel(self.heated_damage_box)
        self.heated_damage_label.setObjectName(u"heated_damage_label")

        self.gridLayout_3.addWidget(self.heated_damage_label, 2, 0, 1, 1)

        self.heated_damage_spin = QDoubleSpinBox(self.heated_damage_box)
        self.heated_damage_spin.setObjectName(u"heated_damage_spin")
        self.heated_damage_spin.setMaximum(100.000000000000000)
        self.heated_damage_spin.setSingleStep(0.100000000000000)
        self.heated_damage_spin.setValue(1.000000000000000)

        self.gridLayout_3.addWidget(self.heated_damage_spin, 2, 1, 1, 1)


        self.scroll_area_layout.addWidget(self.heated_damage_box)

        self.energy_tank_spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.scroll_area_layout.addItem(self.energy_tank_spacer)

        self.scroll_area.setWidget(self.scroll_area_contents)

        self.main_layout.addWidget(self.scroll_area)

        PresetPatcherEnergy.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetPatcherEnergy)

        QMetaObject.connectSlotsByName(PresetPatcherEnergy)
    # setupUi

    def retranslateUi(self, PresetPatcherEnergy):
        PresetPatcherEnergy.setWindowTitle(QCoreApplication.translate("PresetPatcherEnergy", u"Energy", None))
        self.energy_tank_box.setTitle(QCoreApplication.translate("PresetPatcherEnergy", u"Energy tank", None))
        self.dangerous_tank_check.setText(QCoreApplication.translate("PresetPatcherEnergy", u"1 HP mode. In this mode, Energy Tanks and Save Stations leave you at 1 HP instead", None))
        self.energy_tank_capacity_label.setText(QCoreApplication.translate("PresetPatcherEnergy", u"Energy per tank", None))
        self.progressive_damage_reduction_check.setText(QCoreApplication.translate("PresetPatcherEnergy", u"Damage reduction depends on how many suits you have, not which", None))
        self.energy_tank_capacity_spin_box.setSuffix(QCoreApplication.translate("PresetPatcherEnergy", u" energy", None))
        self.energy_tank_capacity_description.setText(QCoreApplication.translate("PresetPatcherEnergy", u"<html><head/><body><p>Configure how much energy each Energy Tank you collect will provide. Your base energy is always this quantity, minus 1.</p><p>While logic will respect this value, only the original value (100) has been tested.</p></body></html>", None))
        self.safe_zone_box.setTitle(QCoreApplication.translate("PresetPatcherEnergy", u"Safe zone", None))
        self.safe_zone_regen_spin.setSuffix(QCoreApplication.translate("PresetPatcherEnergy", u" energy/s", None))
        self.safe_zone_regen_label.setText(QCoreApplication.translate("PresetPatcherEnergy", u"Safe Zone healing rate", None))
        self.safe_zone_logic_heal_check.setText(QCoreApplication.translate("PresetPatcherEnergy", u"Logic considers fully healing at every safe zone. This is currently always on.", None))
        self.safe_zone_description.setText(QCoreApplication.translate("PresetPatcherEnergy", u"<html><head/><body><p>Configure how Dark Aether safe zones operate and how logic uses them.</p></body></html>", None))
        self.dark_aether_box.setTitle(QCoreApplication.translate("PresetPatcherEnergy", u"Dark Aether", None))
        self.varia_suit_spin_box.setSuffix(QCoreApplication.translate("PresetPatcherEnergy", u" energy/s", None))
        self.varia_suit_label.setText(QCoreApplication.translate("PresetPatcherEnergy", u"Varia Suit", None))
        self.dark_suit_label.setText(QCoreApplication.translate("PresetPatcherEnergy", u"Dark Suit", None))
        self.dark_suit_spin_box.setSuffix(QCoreApplication.translate("PresetPatcherEnergy", u" energy/s", None))
        self.dark_aether_label.setText(QCoreApplication.translate("PresetPatcherEnergy", u"<html><head/><body><p>Configure how much damage per second you take in Dark Aether, per suit.<br/>Light Suit is always imune.</p></body></html>", None))
        self.heated_damage_box.setTitle(QCoreApplication.translate("PresetPatcherEnergy", u"Heated rooms", None))
        self.heated_damage_description.setText(QCoreApplication.translate("PresetPatcherEnergy", u"<html><head/><body><p>Configure how much damage per second you take in heated rooms, when you don't have proper protection.</p></body></html>", None))
        self.heated_damage_varia_check.setText(QCoreApplication.translate("PresetPatcherEnergy", u"Only Varia Suit provides heat protection, instead of all suits.", None))
        self.heated_damage_label.setText(QCoreApplication.translate("PresetPatcherEnergy", u"Heated damage rate", None))
        self.heated_damage_spin.setSuffix(QCoreApplication.translate("PresetPatcherEnergy", u" energy/s", None))
    # retranslateUi




import copy
from pathlib import Path
from typing import TypeVar, BinaryIO, Dict, Any

import construct
from construct import (Struct, Int32ub, Const, CString, Byte, Rebuild, Float32b, Flag,
                       Short, PrefixedArray, Array, Switch, If, VarInt, Sequence, Float64b)

from randovania.game_description.world.node import LoreType
from randovania.games.game import RandovaniaGame

X = TypeVar('X')
current_format_version = 8

_IMPOSSIBLE_SET = {"type": "or", "data": {"comment": None, "items": []}}


def convert_to_raw_python(value) -> Any:
    if isinstance(value, construct.ListContainer):
        return [
            convert_to_raw_python(item)
            for item in value
        ]

    if isinstance(value, construct.Container):
        return {
            key: convert_to_raw_python(item)
            for key, item in value.items()
            if not key.startswith("_")
        }

    if isinstance(value, construct.EnumIntegerString):
        return str(value)

    return value


def decode(binary_io: BinaryIO) -> Dict:
    decoded = convert_to_raw_python(ConstructGame.parse_stream(binary_io))

    decoded.pop("format_version")
    decoded.pop("magic_number")
    decoded["initial_states"] = dict(decoded["initial_states"])
    decoded["resource_database"]["requirement_template"] = dict(decoded["resource_database"]["requirement_template"])

    for world in decoded["worlds"]:
        for area in world["areas"]:

            for node in area["nodes"]:
                data = node.pop("data")
                if data is not None:
                    for key, value in data.items():
                        node[key] = value
                node["connections"] = {}

            for i, connections in enumerate(area.pop("connections")):
                node = area["nodes"][i]

                j = 0
                for connection in connections:
                    if j == i:
                        j += 1

                    if connection != _IMPOSSIBLE_SET:
                        node["connections"][area["nodes"][j]["name"]] = connection
                    j += 1

    fields = [
        "game",
        "resource_database",
        "starting_location",
        "initial_states",
        "minimal_logic",
        "victory_condition",
        "dock_weakness_database",
        "worlds",
    ]
    result = {
        field_name: decoded.pop(field_name)
        for field_name in fields
    }
    if decoded:
        raise ValueError(f"Unexpected fields remaining in data: {list(decoded.keys())}")

    return result


def decode_file_path(binary_file_path: Path) -> Dict:
    with binary_file_path.open("rb") as binary_io:
        return decode(binary_io)


def encode(original_data: Dict, x: BinaryIO) -> None:
    data = copy.deepcopy(original_data)

    for world in data["worlds"]:
        for area in world["areas"]:
            area["connections"] = [
                [
                    origin["connections"].get(target["name"], _IMPOSSIBLE_SET)
                    for target in area["nodes"]
                    if origin != target
                ]
                for origin in area["nodes"]
            ]

            for i, node in enumerate(area["nodes"]):
                node.pop("connections")
                area["nodes"][i] = {
                    "name": node.pop("name"),
                    "heal": node.pop("heal"),
                    "coordinates": node.pop("coordinates"),
                    "node_type": node.pop("node_type"),
                    "data": node,
                }

    data["resource_database"]["requirement_template"] = list(data["resource_database"]["requirement_template"].items())
    data["initial_states"] = list(data["initial_states"].items())

    ConstructGame.build_stream(data, x)

    # Resource Info database
    data.pop("game")
    data.pop("resource_database")
    data.pop("dock_weakness_database")
    data.pop("worlds")
    data.pop("victory_condition")
    data.pop("starting_location")
    data.pop("initial_states")
    data.pop("minimal_logic")

    if data:
        raise ValueError(f"Unexpected fields remaining in data: {list(data.keys())}")


def OptionalValue(subcon):
    return construct.FocusedSeq(
        "value",
        present=Rebuild(Flag, construct.this.value != None),
        value=If(construct.this.present, subcon),
    )


def _build_resource_info(**kwargs):
    return Struct(
        index=VarInt,
        long_name=CString("utf8"),
        short_name=CString("utf8"),
        **kwargs,
    )


ConstructResourceInfo = _build_resource_info()

ConstructItemResourceInfo = _build_resource_info(
    max_capacity=Int32ub,
    extra=OptionalValue(Int32ub),
)

ConstructTrickResourceInfo = _build_resource_info(
    description=CString("utf8"),
)

ConstructDamageReductions = Struct(
    index=VarInt,
    reductions=PrefixedArray(VarInt, Struct(
        index=VarInt,
        multiplier=Float32b,
    ))
)

ConstructResourceRequirement = Struct(
    type=Byte,
    index=VarInt,
    amount=Short,
    negate=Flag,
)

requirement_type_map = {
    "resource": ConstructResourceRequirement,
    "template": CString("utf8"),
}

ConstructRequirement = Struct(
    type=construct.Enum(Byte, resource=0, **{"and": 1, "or": 2}, template=3),
    data=Switch(lambda this: this.type, requirement_type_map)
)
ConstructRequirementArray = Struct(
    comment=OptionalValue(CString("utf8")),
    items=PrefixedArray(VarInt, ConstructRequirement),
)

requirement_type_map["and"] = ConstructRequirementArray
requirement_type_map["or"] = ConstructRequirementArray

ConstructDockWeakness = Struct(
    index=VarInt,
    name=CString("utf8"),
    lock_type=VarInt,
    requirement=ConstructRequirement,
)

ConstructResourceDatabase = Struct(
    items=PrefixedArray(VarInt, ConstructItemResourceInfo),
    events=PrefixedArray(VarInt, ConstructResourceInfo),
    tricks=PrefixedArray(VarInt, ConstructTrickResourceInfo),
    damage=PrefixedArray(VarInt, ConstructResourceInfo),
    versions=PrefixedArray(VarInt, ConstructResourceInfo),
    misc=PrefixedArray(VarInt, ConstructResourceInfo),
    requirement_template=PrefixedArray(VarInt, Sequence(CString("utf8"), ConstructRequirement)),
    damage_reductions=PrefixedArray(VarInt, ConstructDamageReductions),
    energy_tank_item_index=VarInt,
    item_percentage_index=OptionalValue(VarInt),
    multiworld_magic_item_index=OptionalValue(VarInt),
)

ConstructResourceGain = Struct(
    resource_type=Byte,
    resource_index=VarInt,
    amount=VarInt,
)

ConstructLoreType = construct.Enum(Byte, **{enum_value.value: i for i, enum_value in enumerate(LoreType)})

ConstructNodeCoordinates = Struct(
    x=Float64b,
    y=Float64b,
    z=Float64b,
)

ConstructNode = Struct(
    name=CString("utf8"),
    heal=Flag,
    coordinates=OptionalValue(ConstructNodeCoordinates),
    node_type=construct.Enum(Byte, generic=0, dock=1, pickup=2, teleporter=3, event=4, translator_gate=5,
                             logbook=6, player_ship=7),
    data=Switch(
        lambda this: this.node_type,
        {
            "dock": Struct(
                dock_index=Byte,
                connected_area_asset_id=VarInt,
                connected_dock_index=Byte,
                dock_type=Byte,
                dock_weakness_index=VarInt,
            ),
            "pickup": Struct(
                pickup_index=VarInt,
                major_location=Flag,
            ),
            "teleporter": Struct(
                destination_world_asset_id=VarInt,
                destination_area_asset_id=VarInt,
                teleporter_instance_id=OptionalValue(VarInt),
                scan_asset_id=OptionalValue(VarInt),
                keep_name_when_vanilla=Flag,
                editable=Flag,
            ),
            "event": Struct(
                event_index=VarInt,
            ),
            "translator_gate": Struct(
                gate_index=VarInt,
            ),
            "logbook": Struct(
                string_asset_id=VarInt,
                lore_type=ConstructLoreType,
                extra=VarInt,
            ),
            "player_ship": Struct(
                is_unlocked=ConstructRequirement,
            )
        }
    )
)

ConstructArea = Struct(
    name=CString("utf8"),
    in_dark_aether=Flag,
    asset_id=VarInt,
    _node_count=Rebuild(VarInt, construct.len_(construct.this.nodes)),
    default_node_index=OptionalValue(VarInt),
    valid_starting_location=Flag,
    nodes=Array(lambda this: this._node_count, ConstructNode),
    connections=Array(
        lambda this: this._node_count,
        Array(lambda this: this._node_count - 1, ConstructRequirement)
    )
)

ConstructWorld = Struct(
    name=CString("utf8"),
    dark_name=OptionalValue(CString("utf8")),
    asset_id=VarInt,
    areas=PrefixedArray(VarInt, ConstructArea),
)

ConstructGameEnum = construct.Enum(Byte, **{enum_item.value: i for i, enum_item in enumerate(RandovaniaGame)})

ConstructMinimalLogicDatabase = Struct(
    items_to_exclude=PrefixedArray(VarInt, Struct(
        index=VarInt,
        when_shuffled=OptionalValue(CString("utf8")),
    )),
    custom_item_amount=PrefixedArray(VarInt, Struct(
        index=VarInt,
        value=VarInt,
    )),
    events_to_exclude=PrefixedArray(VarInt, Struct(
        index=VarInt,
        reason=OptionalValue(CString("utf8")),
    )),
)

ConstructGame = Struct(
    magic_number=Const(b"Req."),
    format_version=Const(current_format_version, Int32ub),
    game=ConstructGameEnum,
    resource_database=ConstructResourceDatabase,
    dock_weakness_database=Struct(
        door=PrefixedArray(VarInt, ConstructDockWeakness),
        portal=PrefixedArray(VarInt, ConstructDockWeakness),
        morph_ball=PrefixedArray(VarInt, ConstructDockWeakness),
    ),
    minimal_logic=OptionalValue(ConstructMinimalLogicDatabase),
    victory_condition=ConstructRequirement,
    starting_location=Struct(
        world_asset_id=VarInt,
        area_asset_id=VarInt,
    ),
    initial_states=PrefixedArray(VarInt, construct.Sequence(CString("utf8"),
                                                            PrefixedArray(VarInt, ConstructResourceGain))),
    worlds=PrefixedArray(VarInt, ConstructWorld),
)



import pytest
from mock import MagicMock, ANY, AsyncMock

from randovania.cli.commands import randomize_command
from randovania.games.game import RandovaniaGame
from randovania.layout.base.cosmetic_patches import BaseCosmeticPatches
from randovania.interface_common.options import Options
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.prime1.prime_cosmetic_patches import PrimeCosmeticPatches
from randovania.layout.prime2.echoes_cosmetic_patches import EchoesCosmeticPatches


@pytest.mark.parametrize(["game", "cosmetic_class"], [
    (RandovaniaGame.METROID_PRIME, PrimeCosmeticPatches),
    (RandovaniaGame.METROID_PRIME_ECHOES, EchoesCosmeticPatches),
])
@pytest.mark.parametrize("with_permalink", [False, True])
def test_randomize_command_logic(mocker, with_permalink, game, cosmetic_class):
    mock_from_str = mocker.patch("randovania.layout.permalink.Permalink.from_str")
    mock_from_file = mocker.patch("randovania.layout.layout_description.LayoutDescription.from_file")
    mock_generate: AsyncMock = mocker.patch("randovania.generator.generator.generate_and_validate_description",
                                            new_callable=AsyncMock)

    mock_provider_for: MagicMock = mocker.patch("randovania.games.patcher_provider.PatcherProvider.patcher_for_game")
    patcher: MagicMock = mock_provider_for.return_value

    args = MagicMock()
    if with_permalink:
        layout_description = mock_generate.return_value
    else:
        layout_description = mock_from_file.return_value
        args.permalink = None

    preset = MagicMock()
    preset.game = game
    layout_description.permalink.player_count = 4
    layout_description.permalink.get_preset = MagicMock(return_value=preset)

    players_config = PlayersConfiguration(
        args.player_index,
        {
            0: "Player 1",
            1: "Player 2",
            2: "Player 3",
            3: "Player 4",
        }
    )

    # Run
    randomize_command.randomize_command_logic(args)

    # Assert
    mock_provider_for.assert_called_once_with(preset.game)

    patcher.create_patch_data.assert_called_once_with(layout_description, players_config, cosmetic_class())
    patcher.patch_game.assert_called_once_with(args.input_file, args.output_file,
                                               patcher.create_patch_data.return_value,
                                               Options.with_default_data_dir().internal_copies_path,
                                               ANY)

    if with_permalink:
        mock_from_str.assert_called_once_with(args.permalink)
        mock_generate.assert_awaited_once_with(permalink=mock_from_str.return_value,
                                               status_update=ANY,
                                               validate_after_generation=True)
        mock_from_file.assert_not_called()
    else:
        mock_from_str.assert_not_called()
        mock_generate.assert_not_called()
        mock_from_file.assert_called_once_with(args.log_file)



import asyncio
import base64
import functools
import hashlib
import logging
import ssl
import time
from enum import Enum
from pathlib import Path
from typing import List, Optional, Tuple

import aiofiles
import aiohttp
import engineio
import socketio
import socketio.exceptions

from randovania.bitpacking import bitpacking
from randovania.game_connection.connection_base import GameConnectionStatus, Inventory
from randovania.game_connection.memory_executor_choice import MemoryExecutorChoice
from randovania.game_description import default_database
from randovania.games.game import RandovaniaGame
from randovania.network_client.game_session import (GameSessionListEntry, GameSessionEntry, User, GameSessionActions,
                                                    GameSessionAction, GameSessionPickups, GameSessionAuditLog,
                                                    GameSessionAuditEntry)
from randovania.network_common import connection_headers
from randovania.network_common.admin_actions import SessionAdminUserAction, SessionAdminGlobalAction
from randovania.network_common.binary_formats import BinaryInventory, BinaryGameSessionActions, \
    BinaryGameSessionAuditLog
from randovania.network_common.error import decode_error, InvalidSession, RequestTimeout, BaseNetworkError
from randovania.network_common.pickup_serializer import BitPackPickupEntry


class ConnectionState(Enum):
    Disconnected = "Disconnected"
    Connecting = "Connecting"
    Connected = "Connected"
    ConnectedNotLogged = "Connected, not logged in"
    ConnectedRestoringSession = "Connected, restoring session"

    @property
    def is_disconnected(self) -> bool:
        return self in (ConnectionState.Disconnected, ConnectionState.Connecting)


def _hash_address(server_address: str) -> str:
    return base64.urlsafe_b64encode(hashlib.blake2b(server_address.encode("utf-8"),
                                                    digest_size=12).digest()).decode("utf-8")


def _decode_pickup(d: str, resource_database):
    decoder = bitpacking.BitPackDecoder(base64.b85decode(d))
    return BitPackPickupEntry.bit_pack_unpack(decoder, resource_database)


class UnableToConnect(Exception):
    def __init__(self, reason: str):
        super().__init__(reason)
        self.reason = reason


_MINIMUM_TIMEOUT = 30
_MAXIMUM_TIMEOUT = 180
_TIMEOUT_STEP = 10

isrgrootx1 = """-----BEGIN CERTIFICATE-----
MIIFazCCA1OgAwIBAgIRAIIQz7DSQONZRGPgu2OCiwAwDQYJKoZIhvcNAQELBQAw
TzELMAkGA1UEBhMCVVMxKTAnBgNVBAoTIEludGVybmV0IFNlY3VyaXR5IFJlc2Vh
cmNoIEdyb3VwMRUwEwYDVQQDEwxJU1JHIFJvb3QgWDEwHhcNMTUwNjA0MTEwNDM4
WhcNMzUwNjA0MTEwNDM4WjBPMQswCQYDVQQGEwJVUzEpMCcGA1UEChMgSW50ZXJu
ZXQgU2VjdXJpdHkgUmVzZWFyY2ggR3JvdXAxFTATBgNVBAMTDElTUkcgUm9vdCBY
MTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAK3oJHP0FDfzm54rVygc
h77ct984kIxuPOZXoHj3dcKi/vVqbvYATyjb3miGbESTtrFj/RQSa78f0uoxmyF+
0TM8ukj13Xnfs7j/EvEhmkvBioZxaUpmZmyPfjxwv60pIgbz5MDmgK7iS4+3mX6U
A5/TR5d8mUgjU+g4rk8Kb4Mu0UlXjIB0ttov0DiNewNwIRt18jA8+o+u3dpjq+sW
T8KOEUt+zwvo/7V3LvSye0rgTBIlDHCNAymg4VMk7BPZ7hm/ELNKjD+Jo2FR3qyH
B5T0Y3HsLuJvW5iB4YlcNHlsdu87kGJ55tukmi8mxdAQ4Q7e2RCOFvu396j3x+UC
B5iPNgiV5+I3lg02dZ77DnKxHZu8A/lJBdiB3QW0KtZB6awBdpUKD9jf1b0SHzUv
KBds0pjBqAlkd25HN7rOrFleaJ1/ctaJxQZBKT5ZPt0m9STJEadao0xAH0ahmbWn
OlFuhjuefXKnEgV4We0+UXgVCwOPjdAvBbI+e0ocS3MFEvzG6uBQE3xDk3SzynTn
jh8BCNAw1FtxNrQHusEwMFxIt4I7mKZ9YIqioymCzLq9gwQbooMDQaHWBfEbwrbw
qHyGO0aoSCqI3Haadr8faqU9GY/rOPNk3sgrDQoo//fb4hVC1CLQJ13hef4Y53CI
rU7m2Ys6xt0nUW7/vGT1M0NPAgMBAAGjQjBAMA4GA1UdDwEB/wQEAwIBBjAPBgNV
HRMBAf8EBTADAQH/MB0GA1UdDgQWBBR5tFnme7bl5AFzgAiIyBpY9umbbjANBgkq
hkiG9w0BAQsFAAOCAgEAVR9YqbyyqFDQDLHYGmkgJykIrGF1XIpu+ILlaS/V9lZL
ubhzEFnTIZd+50xx+7LSYK05qAvqFyFWhfFQDlnrzuBZ6brJFe+GnY+EgPbk6ZGQ
3BebYhtF8GaV0nxvwuo77x/Py9auJ/GpsMiu/X1+mvoiBOv/2X/qkSsisRcOj/KK
NFtY2PwByVS5uCbMiogziUwthDyC3+6WVwW6LLv3xLfHTjuCvjHIInNzktHCgKQ5
ORAzI4JMPJ+GslWYHb4phowim57iaztXOoJwTdwJx4nLCgdNbOhdjsnvzqvHu7Ur
TkXWStAmzOVyyghqpZXjFaH3pO3JLF+l+/+sKAIuvtd7u+Nxe5AW0wdeRlN8NwdC
jNPElpzVmbUq4JUagEiuTDkHzsxHpFKVK7q4+63SM1N95R1NbdWhscdCb+ZAJzVc
oyi3B43njTOQ5yOf+1CceWxG1bQVs5ZufpsMljq4Ui0/1lvh+wjChP4kqKOJ2qxq
4RgqsahDYVvTH9w7jXbyLeiNdd8XM2w9U/t7y0Ff/9yi0GE44Za4rF2LN9d11TPA
mRGunUHBcnWEvgJBQl9nJEiU0Zsnvgc/ubhPgXRR4Xq37Z0j4r7g1SgEEzwxA57d
emyPxgcYxn/eR44/KJ4EBs+lVDR3veyJm+kXQ99b21/+jh5Xos1AnX5iItreGCc=
-----END CERTIFICATE-----"""


class NetworkClient:
    sio: socketio.AsyncClient
    _current_user: Optional[User] = None
    _connection_state: ConnectionState
    _call_lock: asyncio.Lock
    _connect_task: Optional[asyncio.Task] = None
    _waiting_for_on_connect: Optional[asyncio.Future] = None
    _restore_session_task: Optional[asyncio.Task] = None
    _connect_error: Optional[str] = None

    # Game Session
    _current_game_session_meta: Optional[GameSessionEntry] = None
    _current_game_session_actions: Optional[GameSessionActions] = None
    _current_game_session_pickups: Optional[GameSessionPickups] = None
    _current_game_session_audit_log: Optional[GameSessionAuditLog] = None

    def __init__(self, user_data_dir: Path, configuration: dict):
        self.logger = logging.getLogger(__name__)

        old_connect = aiohttp.ClientSession.ws_connect

        @functools.wraps(old_connect)
        def wrap_ws_connect(*args, **kwargs):
            if any("randovania.metroidprime.run" in x for x in args if isinstance(x, str)):
                kwargs["ssl"] = ssl.create_default_context(cadata=isrgrootx1)
            return old_connect(*args, **kwargs)

        aiohttp.ClientSession.ws_connect = wrap_ws_connect

        self._connection_state = ConnectionState.Disconnected
        self.sio = socketio.AsyncClient()
        self._call_lock = asyncio.Lock()
        self._current_timeout = _MINIMUM_TIMEOUT

        self.configuration = configuration
        encoded_address = _hash_address(self.configuration["server_address"])
        self.server_data_path = user_data_dir / encoded_address
        self.session_data_path = self.server_data_path / "session_persistence.bin"

        self.sio.on('connect', self.on_connect)
        self.sio.on('connect_error', self.on_connect_error)
        self.sio.on('disconnect', self.on_disconnect)
        self.sio.on('user_session_update', self.on_user_session_updated)
        self.sio.on("game_session_meta_update", self._on_game_session_meta_update_raw)
        self.sio.on("game_session_actions_update", self._on_game_session_actions_update_raw)
        self.sio.on("game_session_pickups_update", self._on_game_session_pickups_update_raw)
        self.sio.on("game_session_audit_update", self._on_game_session_audit_update_raw)

    @property
    def connection_state(self) -> ConnectionState:
        return self._connection_state

    @connection_state.setter
    def connection_state(self, value: ConnectionState):
        self.logger.debug(f"updated connection_state: {value.value}")
        self._connection_state = value

    async def read_persisted_session(self) -> Optional[bytes]:
        try:
            async with aiofiles.open(self.session_data_path, "rb") as open_file:
                return await open_file.read()

        except FileNotFoundError:
            return None

    def has_previous_session(self) -> bool:
        return self.session_data_path.is_file()

    async def connect_if_authenticated(self):
        if self.session_data_path.is_file():
            self.logger.debug("session data exists")
            try:
                return await self.connect_to_server()
            except UnableToConnect:
                pass
        else:
            self.logger.debug("no session data")

    async def _internal_connect_to_server(self):
        import aiohttp.client_exceptions

        if self.sio.connected:
            return

        waiting_for_on_connect = asyncio.get_running_loop().create_future()
        self._waiting_for_on_connect = waiting_for_on_connect
        try:
            # sio.connect is raising a NotImplementedError, likely due to Windows and/or qasync?
            engineio.asyncio_client.async_signal_handler_set = True

            self.connection_state = ConnectionState.Connecting
            # self.logger.info(f"connect_to_server: sleeping")
            # await asyncio.sleep(1)
            # self.logger.info(f"connect_to_server: sleep over")

            self.logger.info(f"connecting to {self.configuration['server_address']}")
            self._connect_error = None
            await self.sio.connect(
                self.configuration["server_address"],
                socketio_path=self.configuration["socketio_path"],
                transports=['websocket'],
                headers=connection_headers(),
            )
            await waiting_for_on_connect
            self.logger.info(f"connected")

        except (socketio.exceptions.ConnectionError, aiohttp.client_exceptions.ContentTypeError) as e:
            self.logger.info(f"failed with {e} - {type(e)}")
            if self._connect_error is None:
                if isinstance(e, aiohttp.client_exceptions.ContentTypeError):
                    message = e.message
                else:
                    message = str(e)
                await self.on_connect_error(message)
            error = self._connect_error
            await self.sio.disconnect()
            raise UnableToConnect(error)

    def notify_on_connect(self, error_message: Optional[Exception]):
        if self._waiting_for_on_connect is not None:
            if error_message is None:
                self._waiting_for_on_connect.set_result(None)
            else:
                self._waiting_for_on_connect.set_exception(error_message)
            self._waiting_for_on_connect = None

    def connect_to_server(self) -> asyncio.Task:
        if self._connect_task is None:
            self._connect_task = asyncio.create_task(self._internal_connect_to_server())
            self._connect_task.add_done_callback(lambda _: setattr(self, "_connect_task", None))

        return self._connect_task

    async def disconnect_from_server(self):
        self.logger.debug(f"will disconnect")
        await self.sio.disconnect()
        self.logger.debug(f"disconnected")

    async def _restore_session(self):
        persisted_session = await self.read_persisted_session()
        if persisted_session is not None:
            if self._current_game_session_meta is not None:
                session_id = self._current_game_session_meta.id
            else:
                session_id = None
            try:
                self.connection_state = ConnectionState.ConnectedRestoringSession
                self.logger.debug(f"session restoring session, with id {session_id}")
                await self.on_user_session_updated(await self._emit_with_result("restore_user_session",
                                                                                (persisted_session, session_id)))

                if self._current_game_session_meta is not None:
                    await self.game_session_request_update()

                self.logger.info(f"session restored successful")
                self.connection_state = ConnectionState.Connected

            except InvalidSession:
                self.logger.info(f"invalid session, deleting")
                self.connection_state = ConnectionState.ConnectedNotLogged
                self.session_data_path.unlink()
        else:
            self.logger.info(f"no session to restore")
            self.connection_state = ConnectionState.ConnectedNotLogged

    async def on_connect(self):
        error_message = None
        try:
            self._restore_session_task = asyncio.create_task(self._restore_session())
            self._restore_session_task.add_done_callback(lambda _: setattr(self, "_restore_session_task", None))
            await self._restore_session_task

        except BaseNetworkError as e:
            self.logger.warning(f"Unable to restore session after logging in, give up! Reason: {e}")
            error_message = e
            self.connection_state = ConnectionState.Disconnected
            await self.disconnect_from_server()

        finally:
            self.notify_on_connect(error_message)

    async def on_connect_error(self, error_message: str):
        if isinstance(error_message, dict) and "message" in error_message:
            error_message = error_message["message"]

        try:
            self._connect_error = error_message
            self.logger.warning(error_message)
            self.connection_state = ConnectionState.Disconnected
            if self._restore_session_task is not None:
                self._restore_session_task.cancel()
        finally:
            self.notify_on_connect(socketio.exceptions.ConnectionError(error_message))

    async def on_disconnect(self):
        self.logger.info(f"on_disconnect")
        self.connection_state = ConnectionState.Disconnected
        if self._restore_session_task is not None:
            self._restore_session_task.cancel()

    async def on_user_session_updated(self, new_session: dict):
        self._current_user = User.from_json(new_session["user"])
        if self.connection_state in (ConnectionState.ConnectedRestoringSession, ConnectionState.ConnectedNotLogged):
            self.connection_state = ConnectionState.Connected

        self.logger.info(f"{self._current_user.name}, state: {self.connection_state}")

        encoded_session_data = base64.b85decode(new_session["encoded_session_b85"])
        self.server_data_path.mkdir(exist_ok=True, parents=True)
        async with aiofiles.open(self.session_data_path, "wb") as open_file:
            await open_file.write(encoded_session_data)

    # Game Session Updated

    async def _on_game_session_meta_update_raw(self, data: bytes):
        entry = GameSessionEntry.from_json(data)
        self.logger.debug("%s: %s",
                          entry.id,
                          hashlib.blake2b(str(data).encode("utf-8")).hexdigest())
        await self.on_game_session_meta_update(entry)

    async def on_game_session_meta_update(self, entry: GameSessionEntry):
        self._current_game_session_meta = entry

    async def _on_game_session_actions_update_raw(self, data: bytes):
        await self.on_game_session_actions_update(GameSessionActions(
            tuple(GameSessionAction.from_json(item)
                  for item in BinaryGameSessionActions.parse(data))
        ))

    async def on_game_session_actions_update(self, actions: GameSessionActions):
        self._current_game_session_actions = actions

    async def _on_game_session_pickups_update_raw(self, data):
        game = RandovaniaGame(data["game"])
        resource_database = default_database.resource_database_for(game)

        await self.on_game_session_pickups_update(GameSessionPickups(
            game=game,
            pickups=tuple(
                (item["provider_name"], _decode_pickup(item["pickup"], resource_database))
                for item in data["pickups"]
            ),
        ))

    async def on_game_session_pickups_update(self, pickups: GameSessionPickups):
        self._current_game_session_pickups = pickups

    async def _on_game_session_audit_update_raw(self, data):
        await self.on_game_session_audit_update(GameSessionAuditLog(
            entries=tuple(
                GameSessionAuditEntry.from_json(entry)
                for entry in BinaryGameSessionAuditLog.parse(data)
            ),
        ))

    async def on_game_session_audit_update(self, audit_log: GameSessionAuditLog):
        self._current_game_session_audit_log = audit_log

    #

    def _update_timeout_with(self, request_time: float, success: bool):
        if success:
            if request_time < self._current_timeout - _TIMEOUT_STEP and self._current_timeout > _MINIMUM_TIMEOUT:
                self._current_timeout -= _TIMEOUT_STEP
                self.logger.debug(f"decreasing timeout by {_TIMEOUT_STEP}, to {self._current_timeout}")
        else:
            self._current_timeout += _TIMEOUT_STEP
            self.logger.debug(f"increasing timeout by {_TIMEOUT_STEP}, to {self._current_timeout}")

        self._current_timeout = min(max(self._current_timeout, _MINIMUM_TIMEOUT), _MAXIMUM_TIMEOUT)

    async def _emit_without_result(self, event, data=None, namespace=None):
        if self.connection_state.is_disconnected:
            self.logger.debug(f"{event}, urgent connect start")
            await self.connect_to_server()
            self.logger.debug(f"{event}, urgent connect finished")

        self.logger.debug(f"{event}, getting lock")
        async with self._call_lock:
            self.logger.debug(f"{event}, will emit")
            await self.sio.emit(event, data, namespace=namespace)

    async def _emit_with_result(self, event, data=None, namespace=None):
        if self.connection_state.is_disconnected:
            self.logger.debug(f"{event}, urgent connect start")
            await self.connect_to_server()
            self.logger.debug(f"{event}, urgent connect finished")

        self.logger.debug(f"{event}, getting lock")
        async with self._call_lock:
            request_start = time.time()
            timeout = self._current_timeout
            self.logger.debug(f"{event}, will call with timeout {timeout}")
            try:
                result = await self.sio.call(event, data, namespace=namespace, timeout=timeout)
                request_time = time.time() - request_start
                self._update_timeout_with(request_time, True)

            except socketio.exceptions.TimeoutError:
                request_time = time.time() - request_start
                self._update_timeout_with(request_time, False)
                raise RequestTimeout(f"Timeout after {request_time:.2f}s, with a timeout of {timeout}.")

        if result is None:
            return None

        possible_error = decode_error(result)
        if possible_error is None:
            return result["result"]
        else:
            raise possible_error

    async def game_session_request_update(self):
        await self._emit_without_result("game_session_request_update", self._current_game_session_meta.id)

    async def game_session_collect_locations(self, locations: Tuple[int, ...]):
        await self._emit_with_result("game_session_collect_locations",
                                     (self._current_game_session_meta.id, locations))

    async def get_game_session_list(self) -> List[GameSessionListEntry]:
        return [
            GameSessionListEntry.from_json(item)
            for item in await self._emit_with_result("list_game_sessions")
        ]

    async def create_new_session(self, session_name: str) -> GameSessionEntry:
        result = await self._emit_with_result("create_game_session", session_name)
        self._current_game_session_meta = GameSessionEntry.from_json(result)
        return self._current_game_session_meta

    async def join_game_session(self, session: GameSessionListEntry, password: Optional[str]):
        result = await self._emit_with_result("join_game_session", (session.id, password))
        self._current_game_session_meta = GameSessionEntry.from_json(result)

    async def leave_game_session(self, permanent: bool):
        if permanent:
            await self.session_admin_player(self._current_user.id, SessionAdminUserAction.KICK, None)
        await self._emit_with_result("disconnect_game_session", self._current_game_session_meta.id)
        self._current_game_session_meta = None

    async def session_admin_global(self, action: SessionAdminGlobalAction, arg):
        return await self._emit_with_result("game_session_admin_session",
                                            (self._current_game_session_meta.id, action.value, arg))

    async def session_admin_player(self, user_id: int, action: SessionAdminUserAction, arg):
        return await self._emit_with_result("game_session_admin_player",
                                            (self._current_game_session_meta.id, user_id, action.value, arg))

    async def session_self_update(self, inventory: Inventory, state: GameConnectionStatus,
                                  backend: MemoryExecutorChoice):

        inventory_binary = BinaryInventory.build([
            {"index": resource.index, "amount": item.amount, "capacity": item.capacity}
            for resource, item in inventory.items()
        ])
        state_string = f"{state.pretty_text} ({backend.pretty_text})"

        await self._emit_without_result("game_session_self_update",
                                        (self._current_game_session_meta.id, inventory_binary, state_string))

    @property
    def current_user(self) -> Optional[User]:
        return self._current_user

    @property
    def current_game_session(self) -> Optional[GameSessionEntry]:
        return self._current_game_session_meta



from randovania.layout.game_to_class import AnyGameConfiguration
from typing import Tuple
from PySide2 import QtWidgets

from PySide2.QtCore import Signal
from PySide2.QtWidgets import QDialog, QWidget

from randovania.gui.lib.common_qt_lib import set_default_window_icon
from randovania.layout.base.major_item_state import MajorItemState
from randovania.gui.generated.widget_pickup_style_ui import Ui_PickupStyleWidget
from randovania.interface_common.preset_editor import PresetEditor
from randovania.gui.lib import signal_handling
from randovania.layout.base.pickup_model import PickupModelDataSource, PickupModelStyle
from randovania.gui.lib.scroll_protected import ScrollProtectedComboBox

class PickupStyleWidget(QDialog, Ui_PickupStyleWidget):
    Changed = Signal()
    
    def __init__(self, parent: QWidget, editor: PresetEditor):
        super().__init__(parent)
        self.setupUi(self)
        self._editor = editor

        # Item Data
        for i, value in enumerate(PickupModelStyle):
            self.pickup_model_combo.setItemData(i, value)
        for i, value in enumerate(PickupModelDataSource):
            self.pickup_data_source_combo.setItemData(i, value)

        # TODO: implement the LOCATION data source
        self.pickup_data_source_combo.removeItem(self.pickup_data_source_combo.findData(PickupModelDataSource.LOCATION))
        self.pickup_model_combo.currentIndexChanged.connect(
            self._persist_enum(self.pickup_model_combo, "pickup_model_style"))
        self.pickup_data_source_combo.currentIndexChanged.connect(
            self._persist_enum(self.pickup_data_source_combo, "pickup_model_data_source"))

    def _persist_enum(self, combo: QtWidgets.QComboBox, attribute_name: str):
        def persist(index: int):
            with self._editor as options:
                options.set_configuration_field(attribute_name, combo.itemData(index))
        return persist

    def update(self, layout: AnyGameConfiguration):
        self.pickup_model_combo.setCurrentIndex(self.pickup_model_combo.findData(layout.pickup_model_style))
        self.pickup_data_source_combo.setCurrentIndex(
            self.pickup_data_source_combo.findData(layout.pickup_model_data_source))
        self.pickup_data_source_combo.setEnabled(layout.pickup_model_style != PickupModelStyle.ALL_VISIBLE)




import collections
import dataclasses
from functools import partial
from typing import Dict, Tuple, List

from PySide2 import QtWidgets, QtCore

from randovania.game_description import default_database
from randovania.game_description.item.ammo import Ammo
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.item.item_database import ItemDatabase
from randovania.game_description.item.major_item import MajorItem
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_type import ResourceType
from randovania.games.game import RandovaniaGame
from randovania.games.prime.patcher_file_lib import item_names
from randovania.generator.item_pool import pool_creator
from randovania.gui.generated.preset_item_pool_ui import Ui_PresetItemPool
from randovania.gui.lib import common_qt_lib
from randovania.gui.lib.foldable import Foldable
from randovania.gui.lib.scroll_protected import ScrollProtectedComboBox, ScrollProtectedSpinBox
from randovania.gui.preset_settings.item_configuration_widget import ItemConfigurationWidget
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.gui.preset_settings.split_ammo_widget import AmmoPickupWidgets
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.base.major_item_state import MajorItemState
from randovania.layout.preset import Preset
from randovania.resolver.exceptions import InvalidConfiguration

_EXPECTED_COUNT_TEXT_TEMPLATE_EXACT = ("A total of {total} will be available."
                                       "\n{from_items} will be provided from major items."
                                       "\n{maximum} is the maximum you can have at once.")


class PresetItemPool(PresetTab, Ui_PresetItemPool):
    game: RandovaniaGame
    _boxes_for_category: Dict[str, Tuple[QtWidgets.QGroupBox, QtWidgets.QGridLayout,
                                         Dict[MajorItem, ItemConfigurationWidget]]]
    _default_items: Dict[ItemCategory, QtWidgets.QComboBox]

    _ammo_item_count_spinboxes: Dict[str, List[QtWidgets.QSpinBox]]
    _ammo_pickup_widgets: Dict[Ammo, AmmoPickupWidgets]

    def __init__(self, editor: PresetEditor):
        super().__init__(editor)
        self.setupUi(self)

        size_policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        self.item_pool_layout.setAlignment(QtCore.Qt.AlignTop)

        # Relevant Items
        self.game = editor.game
        self.game_description = default_database.game_description_for(self.game)
        item_database = default_database.item_database_for_game(self.game)

        self._register_random_starting_events()
        self._create_categories_boxes(item_database, size_policy)
        self._create_customizable_default_items(item_database)
        self._create_major_item_boxes(item_database, self.game_description.resource_database)
        self._create_ammo_pickup_boxes(size_policy, item_database)

    @property
    def uses_patches_tab(self) -> bool:
        return False

    def on_preset_changed(self, preset: Preset):
        layout = preset.configuration
        major_configuration = layout.major_items_configuration

        # General
        self.multi_pickup_placement_check.setChecked(layout.multi_pickup_placement)

        # Random Starting Items
        self.minimum_starting_spinbox.setValue(major_configuration.minimum_random_starting_items)
        self.maximum_starting_spinbox.setValue(major_configuration.maximum_random_starting_items)

        # Default Items
        for category, default_item in major_configuration.default_items.items():
            combo = self._default_items[category]
            combo.setCurrentIndex(combo.findData(default_item))

            for item, widget in self._boxes_for_category[category.name][2].items():
                widget.setEnabled(default_item != item)

        # Major Items
        for _, _, elements in self._boxes_for_category.values():
            for major_item, widget in elements.items():
                widget.state = major_configuration.items_state[major_item]

        # Ammo
        ammo_provided = major_configuration.calculate_provided_ammo()
        ammo_configuration = layout.ammo_configuration

        resource_database = self.game_description.resource_database

        item_for_index: Dict[int, ItemResourceInfo] = {
            ammo_index: resource_database.get_item(ammo_index)
            for ammo_index in ammo_provided.keys()
        }

        for ammo, state in ammo_configuration.items_state.items():
            widgets = self._ammo_pickup_widgets[ammo]
            widgets.pickup_spinbox.setValue(state.pickup_count)

            if widgets.require_major_item_check is not None:
                widgets.require_major_item_check.setChecked(state.requires_major_item)

            totals = []
            for ammo_index, count in enumerate(state.ammo_count):
                totals.append(ammo_provided[ammo.items[ammo_index]] + count * state.pickup_count)
                self._ammo_item_count_spinboxes[ammo.name][ammo_index].setValue(count)

            try:
                if state.pickup_count == 0:
                    widgets.expected_count.setText("No expansions will be created.")
                    continue

                widgets.expected_count.setText(
                    _EXPECTED_COUNT_TEXT_TEMPLATE_EXACT.format(
                        total=" and ".join(
                            item_names.add_quantity_to_resource(item_for_index[ammo_index].long_name,
                                                                total, True)
                            for ammo_index, total in zip(ammo.items, totals)
                        ),
                        from_items=" and ".join(
                            item_names.add_quantity_to_resource(item_for_index[ammo_index].long_name,
                                                                ammo_provided[ammo_index], True)
                            for ammo_index in ammo.items
                        ),
                        maximum=" and ".join(
                            item_names.add_quantity_to_resource(item_for_index[ammo_index].long_name,
                                                                item_for_index[ammo_index].max_capacity, True)
                            for ammo_index in ammo.items
                        )
                    )
                )

            except InvalidConfiguration as invalid_config:
                widgets.expected_count.setText(str(invalid_config))

        # Item pool count
        try:
            pool_items, maximum_size = pool_creator.calculate_pool_item_count(layout)
            self.item_pool_count_label.setText("Items in pool: {}/{}".format(pool_items, maximum_size))
            common_qt_lib.set_error_border_stylesheet(self.item_pool_count_label, pool_items > maximum_size)

        except InvalidConfiguration as invalid_config:
            self.item_pool_count_label.setText("Invalid Configuration: {}".format(invalid_config))
            common_qt_lib.set_error_border_stylesheet(self.item_pool_count_label, True)

    # Random Starting
    def _register_random_starting_events(self):
        self.minimum_starting_spinbox.valueChanged.connect(self._on_update_minimum_starting)
        self.maximum_starting_spinbox.valueChanged.connect(self._on_update_maximum_starting)

    def _on_update_minimum_starting(self, value: int):
        with self._editor as options:
            options.major_items_configuration = dataclasses.replace(options.major_items_configuration,
                                                                    minimum_random_starting_items=value)

    def _on_update_maximum_starting(self, value: int):
        with self._editor as options:
            options.major_items_configuration = dataclasses.replace(options.major_items_configuration,
                                                                    maximum_random_starting_items=value)

    def _create_categories_boxes(self, item_database: ItemDatabase, size_policy):
        self._boxes_for_category = {}

        categories = set()
        for major_item in item_database.major_items.values():
            if not major_item.required:
                categories.add(major_item.item_category)

        all_categories = list(item_database.item_categories.values())
        for major_item_category in sorted(categories, key=lambda it: all_categories.index(it)):
            category_box = Foldable(major_item_category.long_name)
            category_box.setObjectName(f"category_box {major_item_category}")
            category_layout = QtWidgets.QGridLayout()
            category_layout.setObjectName(f"category_layout {major_item_category}")
            category_box.set_content_layout(category_layout)

            self.item_pool_layout.addWidget(category_box)
            self._boxes_for_category[major_item_category.name] = category_box, category_layout, {}

    def _create_customizable_default_items(self, item_database: ItemDatabase):
        self._default_items = {}

        for category, possibilities in item_database.default_items.items():
            parent, layout, _ = self._boxes_for_category[category.name]

            label = QtWidgets.QLabel(parent)
            label.setText(f"Default {category.long_name}")
            layout.addWidget(label, 0, 0)

            combo = ScrollProtectedComboBox(parent)
            for item in possibilities:
                combo.addItem(item.name, item)
            combo.currentIndexChanged.connect(partial(self._on_default_item_updated, category, combo))
            layout.addWidget(combo, 0, 1)

            if len(possibilities) <= 1:
                label.hide()
                combo.hide()

            self._default_items[category] = combo

    def _on_default_item_updated(self, category: ItemCategory, combo: QtWidgets.QComboBox, _):
        item: MajorItem = combo.currentData()
        with self._editor as editor:
            new_config = editor.major_items_configuration
            new_config = new_config.replace_default_item(category, item)
            new_config = new_config.replace_state_for_item(
                item,
                MajorItemState(
                    num_included_in_starting_items=1,
                    included_ammo=new_config.items_state[item].included_ammo
                ),
            )
            editor.major_items_configuration = new_config

    def _create_major_item_boxes(self, item_database: ItemDatabase, resource_database: ResourceDatabase):
        for major_item in item_database.major_items.values():
            if major_item.required or major_item.item_category.name == "energy_tank":
                continue

            category_box, category_layout, elements = self._boxes_for_category[major_item.item_category.name]
            widget = ItemConfigurationWidget(None, major_item, MajorItemState(), resource_database)
            widget.Changed.connect(partial(self._on_major_item_updated, widget))

            row = category_layout.rowCount()
            if row > 1:
                # Show the transparent separator line if it's not the first element
                widget.separator_line.show()

            category_layout.addWidget(widget, row, 0, 1, 2)
            elements[major_item] = widget

    def _on_major_item_updated(self, item_widget: ItemConfigurationWidget):
        with self._editor as editor:
            editor.major_items_configuration = editor.major_items_configuration.replace_state_for_item(
                item_widget.item, item_widget.state
            )

    # Ammo
    def _create_ammo_pickup_boxes(self, size_policy, item_database: ItemDatabase):
        """
        Creates the GroupBox with SpinBoxes for selecting the pickup count of all the ammo
        :param item_database:
        :return:
        """

        self._ammo_item_count_spinboxes = collections.defaultdict(list)
        self._ammo_pickup_widgets = {}

        resource_database = default_database.resource_database_for(self.game)
        broad_to_category = {
            "beam_related": "beam",
            "morph_ball_related": "morph_ball",
            "missile_related": "missile",
        }

        for ammo in item_database.ammo.values():
            category_box, category_layout, _ = self._boxes_for_category[broad_to_category[ammo.broad_category.name]]

            pickup_box = QtWidgets.QGroupBox(category_box)
            pickup_box.setSizePolicy(size_policy)
            pickup_box.setTitle(ammo.name + "s")
            layout = QtWidgets.QGridLayout(pickup_box)
            layout.setObjectName(f"{ammo.name} Box Layout")

            current_row = 0

            def add_row(widget_a: QtWidgets.QWidget, widget_b: QtWidgets.QWidget):
                nonlocal current_row
                layout.addWidget(widget_a, current_row, 0)
                layout.addWidget(widget_b, current_row, 1)
                current_row += 1

            for ammo_index, ammo_item in enumerate(ammo.items):
                item = resource_database.get_by_type_and_index(ResourceType.ITEM, ammo_item)

                item_count_label = QtWidgets.QLabel(pickup_box)
                item_count_label.setText(item.long_name if len(ammo.items) > 1 else "Contains")

                item_count_spinbox = ScrollProtectedSpinBox(pickup_box)
                item_count_spinbox.setMaximum(item.max_capacity)
                item_count_spinbox.valueChanged.connect(partial(self._on_update_ammo_pickup_item_count_spinbox,
                                                                ammo, ammo_index))
                self._ammo_item_count_spinboxes[ammo.name].append(item_count_spinbox)
                add_row(item_count_label, item_count_spinbox)

            # Pickup Count
            count_label = QtWidgets.QLabel(pickup_box)
            count_label.setText("Pickup Count")
            count_label.setToolTip("How many instances of this expansion should be placed.")

            pickup_spinbox = ScrollProtectedSpinBox(pickup_box)
            pickup_spinbox.setMaximum(999)
            pickup_spinbox.valueChanged.connect(partial(self._on_update_ammo_pickup_num_count_spinbox, ammo))

            add_row(count_label, pickup_spinbox)

            if ammo.temporary:
                require_major_item_check = QtWidgets.QCheckBox(pickup_box)
                require_major_item_check.setText("Requires the major item to work?")
                require_major_item_check.stateChanged.connect(partial(self._on_update_ammo_require_major_item, ammo))
                layout.addWidget(require_major_item_check, current_row, 0, 1, 2)
                current_row += 1
            else:
                require_major_item_check = None

            expected_count = QtWidgets.QLabel(pickup_box)
            expected_count.setWordWrap(True)
            expected_count.setText("<TODO>")
            layout.addWidget(expected_count, current_row, 0, 1, 2)
            current_row += 1

            self._ammo_pickup_widgets[ammo] = AmmoPickupWidgets(pickup_spinbox, expected_count,
                                                                pickup_box, require_major_item_check)
            category_layout.addWidget(pickup_box)

    def _on_update_ammo_pickup_item_count_spinbox(self, ammo: Ammo, ammo_index: int, value: int):
        with self._editor as options:
            ammo_configuration = options.ammo_configuration
            state = ammo_configuration.items_state[ammo]
            ammo_count = list(state.ammo_count)
            ammo_count[ammo_index] = value

            options.ammo_configuration = ammo_configuration.replace_state_for_ammo(
                ammo,
                dataclasses.replace(state, ammo_count=tuple(ammo_count))
            )

    def _on_update_ammo_pickup_num_count_spinbox(self, ammo: Ammo, value: int):
        with self._editor as options:
            ammo_configuration = options.ammo_configuration
            options.ammo_configuration = ammo_configuration.replace_state_for_ammo(
                ammo,
                dataclasses.replace(ammo_configuration.items_state[ammo], pickup_count=value)
            )

    def _on_update_ammo_require_major_item(self, ammo: Ammo, value: int):
        with self._editor as options:
            ammo_configuration = options.ammo_configuration
            options.ammo_configuration = ammo_configuration.replace_state_for_ammo(
                ammo,
                dataclasses.replace(ammo_configuration.items_state[ammo], requires_major_item=bool(value))
            )



import dataclasses
from typing import Tuple, Iterator

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder
from randovania.game_description import default_database
from randovania.game_description.item.major_item import MajorItem

ENERGY_TANK_MAXIMUM_COUNT = 16
DEFAULT_MAXIMUM_SHUFFLED = (2, 10, 99)


@dataclasses.dataclass(frozen=True)
class MajorItemState:
    include_copy_in_original_location: bool = False
    num_shuffled_pickups: int = 0
    num_included_in_starting_items: int = 0
    included_ammo: Tuple[int, ...] = tuple()

    def check_consistency(self, item: MajorItem):
        db = default_database.resource_database_for(item.game)

        if item.required:
            if self != MajorItemState(num_included_in_starting_items=1):
                raise ValueError("Required items must be included in starting items.")

        if self.num_included_in_starting_items > 0:
            if len(item.progression) > 1:
                raise ValueError("Progressive items cannot be starting item.")

            for progression in item.progression:
                if self.num_included_in_starting_items > db.get_item(progression).max_capacity:
                    raise ValueError("More starting copies than the item's max copy.")

        if self.include_copy_in_original_location and item.original_index is None:
            raise ValueError("Custom item cannot be vanilla.")

        if len(self.included_ammo) != len(item.ammo_index):
            raise ValueError("Mismatched included_ammo array size")

        for ammo_index, ammo in zip(item.ammo_index, self.included_ammo):
            if ammo > db.get_item(ammo_index).max_capacity:
                raise ValueError(f"Including more than maximum capacity for ammo {ammo_index}")

    @property
    def as_json(self) -> dict:
        result: dict = {}

        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if value != field.default:
                result[field.name] = value

        if "included_ammo" in result:
            result["included_ammo"] = list(result["included_ammo"])

        return result

    @classmethod
    def from_json(cls, value: dict) -> "MajorItemState":
        kwargs: dict = {}

        for field in dataclasses.fields(cls):
            if field.name in value:
                kwargs[field.name] = value[field.name]

        if "included_ammo" in kwargs:
            kwargs["included_ammo"] = tuple(kwargs["included_ammo"])

        return cls(**kwargs)

    def bit_pack_encode(self, item: MajorItem, reference: "MajorItemState") -> Iterator[Tuple[int, int]]:
        db = default_database.resource_database_for(item.game)
        if item.progression:
            main_index = item.progression[0]
        else:
            main_index = item.ammo_index[0]
        main_item = db.get_item(main_index)

        # original location
        if item.original_index is not None:
            yield from bitpacking.encode_bool(self.include_copy_in_original_location)

        # num shuffled
        yield from bitpacking.encode_int_with_limits(self.num_shuffled_pickups, DEFAULT_MAXIMUM_SHUFFLED)

        # starting item
        if main_item.max_capacity > 1:
            yield from bitpacking.encode_int_with_limits(self.num_included_in_starting_items,
                                                         (2, main_item.max_capacity + 1))
        else:
            yield self.num_included_in_starting_items, main_item.max_capacity + 1

        # ammo index
        assert len(self.included_ammo) == len(item.ammo_index)
        if self.included_ammo:
            custom_ammo = self.included_ammo != reference.included_ammo
            yield from bitpacking.encode_bool(custom_ammo)
            if custom_ammo:
                all_equal = len(set(self.included_ammo)) == 1
                if len(item.ammo_index) > 1:
                    yield from bitpacking.encode_bool(all_equal)

                for ammo_index, ammo in zip(item.ammo_index, self.included_ammo):
                    yield ammo, db.get_item(ammo_index).max_capacity + 1
                    if all_equal:
                        break

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, item: MajorItem, reference: "MajorItemState") -> "MajorItemState":
        db = default_database.resource_database_for(item.game)
        if item.progression:
            main_index = item.progression[0]
        else:
            main_index = item.ammo_index[0]
        main_item = db.get_item(main_index)

        # original location
        original = False
        if item.original_index is not None:
            original = bitpacking.decode_bool(decoder)

        # num shuffled
        shuffled = bitpacking.decode_int_with_limits(decoder, DEFAULT_MAXIMUM_SHUFFLED)

        # starting item
        if main_item.max_capacity > 1:
            starting = bitpacking.decode_int_with_limits(decoder, (2, main_item.max_capacity + 1))
        else:
            starting = decoder.decode_single(main_item.max_capacity + 1)

        # ammo index
        if item.ammo_index:
            custom_ammo = bitpacking.decode_bool(decoder)
            if custom_ammo:
                all_equal = len(item.ammo_index) <= 1 or bitpacking.decode_bool(decoder)
                if all_equal:
                    ammo = decoder.decode_single(db.get_item(item.ammo_index[0]).max_capacity + 1)
                    included_ammo = [ammo] * len(item.ammo_index)
                else:
                    included_ammo = [decoder.decode_single(db.get_item(item).max_capacity + 1)
                                     for item in item.ammo_index]
            else:
                included_ammo = reference.included_ammo
        else:
            included_ammo = []

        return cls(
            include_copy_in_original_location=original,
            num_shuffled_pickups=shuffled,
            num_included_in_starting_items=starting,
            included_ammo=tuple(included_ammo),
        )
